from __future__ import annotations

import shiny.experimental as x
from shiny import App, ui

app_ui = ui.page_fluid(
    x.ui.layout_sidebar(
        "Sidebar content",
        "Main content",
    ),
)


app = App(app_ui, server=None)



from __future__ import annotations

from typing import TypeVar

from htmltools import Tag

from ._htmldeps import fill_dependency
from ._tag import tag_prepend_class

__all__ = (
    "as_fillable_container",
    "as_fill_item",
)

TagT = TypeVar("TagT", bound="Tag")


FILL_ITEM_CLASS = "html-fill-item"
FILL_CONTAINER_CLASS = "html-fill-container"


def as_fillable_container(
    tag: TagT,
) -> TagT:
    tag_prepend_class(tag, FILL_CONTAINER_CLASS)
    tag.append(fill_dependency())
    return tag


def as_fill_item(
    tag: TagT,
) -> TagT:
    tag_prepend_class(tag, FILL_ITEM_CLASS)
    tag.append(fill_dependency())
    return tag



from __future__ import annotations

from typing import Literal, Optional, Sequence

from htmltools import (
    MetadataNode,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagList,
    css,
    tags,
)

from ...types import MISSING, MISSING_TYPE, NavSetArg
from ...ui._page import page_bootstrap
from ...ui._utils import get_window_title
from ._css_unit import CssUnit, as_css_padding, as_css_unit
from ._fill import as_fillable_container
from ._htmldeps import page_fillable_dependency, page_sidebar_dependency
from ._navs import navset_bar
from ._sidebar import Sidebar, layout_sidebar
from ._utils import consolidate_attrs


def page_sidebar(
    sidebar: Sidebar | TagChild | TagAttrs,
    *args: TagChild | TagAttrs,
    title: Optional[str | Tag | TagList] = None,
    fillable: bool = True,
    fillable_mobile: bool = False,
    window_title: str | MISSING_TYPE = MISSING,
    lang: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a page with a sidebar and a title.

    Parameters
    ----------
    args
        UI elements.
    sidebar
        Content to display in the sidebar.
    title
        A title to display at the top of the page.
    fillable
        Whether or not the main content area should be considered a fillable
        (i.e., flexbox) container.
    fillable_mobile
        Whether or not ``fillable`` should apply on mobile devices.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    kwargs
        Additional attributes passed to :func:`~shiny.ui.layout_sidebar`.

    Returns
    -------
    :
        A UI element.
    """

    if isinstance(title, str):
        title = tags.h1(title, class_="bslib-page-title")

    attrs, children = consolidate_attrs(*args, **kwargs)

    return page_fillable(
        title,
        layout_sidebar(
            sidebar,
            *children,
            attrs,
            fillable=fillable,
            border=False,
            border_radius=False,
        ),
        get_window_title(title, window_title=window_title),
        page_sidebar_dependency(),
        padding=0,
        gap=0,
        lang=lang,
        fillable_mobile=fillable_mobile,
    )


def page_navbar(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    title: Optional[str | Tag | TagList] = None,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    # Only page_navbar gets enhancedtreatement for `fillable`
    # If an `*args`'s `data-value` attr string is in `fillable`, then the component is fillable
    fillable: bool | list[str] = True,
    fillable_mobile: bool = False,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    position: Literal["static-top", "fixed-top", "fixed-bottom"] = "static-top",
    header: Optional[TagChild] = None,
    footer: Optional[TagChild] = None,
    bg: Optional[str] = None,
    inverse: bool = True,
    collapsible: bool = True,
    fluid: bool = True,
    window_title: str | MISSING_TYPE = MISSING,
    lang: Optional[str] = None,
) -> Tag:
    """
    Create a page with a navbar and a title.

    Parameters
    ----------

    args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :func:`~shiny.ui.sidebar` component to display on every page.
    fillable
        Whether or not the main content area should be considered a fillable
        (i.e., flexbox) container.
    fillable_mobile
        Whether or not ``fillable`` should apply on mobile devices.
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the elements into an expandable menu on mobile devices or narrow window widths.
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.

    Returns
    -------
    :
        A UI element.

    See Also
    -------
    * :func:`~shiny.ui.nav`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.experimental.ui.navset_bar`
    * :func:`~shiny.ui.page_fluid`

    Example
    -------
    See :func:`~shiny.ui.nav`.
    """
    if sidebar is not None and not isinstance(sidebar, Sidebar):
        raise TypeError(
            "`sidebar=` is not a `Sidebar` instance. Use `ui.sidebar(...)` to create one."
        )

    # If a sidebar is provided, we want the layout_sidebar(fill = TRUE) component
    # (which is a sibling of the <nav>) to always fill the page
    if fillable is False and sidebar is None:
        # `page_func = page_bootstrap` throws type errors. Wrap in a function to get around them
        def page_func(*args: TagChild | TagAttrs, **kwargs: TagAttrValue) -> Tag:
            return page_bootstrap(*args, **kwargs)

    else:

        def page_func(*args: TagChild | TagAttrs, **kwargs: TagAttrValue) -> Tag:
            return page_fillable(
                *args,
                fillable_mobile=fillable_mobile,
                padding=0,
                gap=0,
                **kwargs,
            )

    return page_func(
        navset_bar(
            *args,
            title=title,
            id=id,
            selected=selected,
            sidebar=sidebar,
            fillable=fillable,
            gap=gap,
            padding=padding,
            position=position,
            header=header,
            footer=footer,
            bg=bg,
            inverse=inverse,
            collapsible=collapsible,
            fluid=fluid,
        ),
        get_window_title(title, window_title=window_title),
        title=None,
        # theme = theme,
        lang=lang,
    )


def page_fillable(
    *args: TagChild | TagAttrs,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fillable_mobile: bool = False,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Creates a fillable page

    Parameters
    ----------
    *args
        UI elements.
    padding
        Padding to use for the body. See :func:`~shiny.experimental.ui.as_css_padding`
        for more details.
    fillable_mobile
        Whether or not the page should fill the viewport's height on mobile devices
        (i.e., narrow windows).
    gap
        A CSS length unit passed through :func:`~shiny.experimental.ui.as_css_unit`
        defining the `gap` (i.e., spacing) between elements provided to `*args`.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.

    Returns
    -------
    :
        A UI element.

    See Also
    -------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_fixed`
    """
    attrs, children = consolidate_attrs(*args, **kwargs)

    style = css(
        padding=as_css_padding(padding),
        gap=as_css_unit(gap),
        __bslib_page_fill_mobile_height="100%" if fillable_mobile else "auto",
    )

    return page_bootstrap(
        tags.head(tags.style("html { height: 100%; }")),
        as_fillable_container(
            tags.body(
                {"class": "bslib-page-fill bslib-gap-spacing", "style": style},
                attrs,
                *children,
            )
        ),
        page_fillable_dependency(),
        title=title,
        lang=lang,
    )



from ._accordion import (
    AccordionPanel,
    accordion,
    accordion_panel,
    accordion_panel_close,
    accordion_panel_insert,
    accordion_panel_open,
    accordion_panel_remove,
    accordion_panel_set,
    update_accordion_panel,
)
from ._card import (
    CardItem,
    ImgContainer,
    TagCallable,
    WrapperCallable,
    card,
    card_body,
    card_footer,
    card_header,
    card_image,
    card_title,
)
from ._css_unit import CssUnit, as_css_padding, as_css_unit
from ._fill import (  # bind_fill_role,
    FillingLayout,
    as_fill_carrier,
    as_fill_item,
    as_fillable_container,
    is_fill_carrier,
    is_fill_item,
    is_fillable_container,
    remove_all_fill,
)
from ._input_text import input_text_area
from ._layout import layout_column_wrap
from ._navs import navset_bar, navset_pill_card, navset_tab_card
from ._output import output_image, output_plot, output_ui
from ._page import page_fillable, page_navbar, page_sidebar
from ._sidebar import (
    DeprecatedPanelMain,
    DeprecatedPanelSidebar,
    Sidebar,
    layout_sidebar,
    panel_main,
    panel_sidebar,
    sidebar,
    sidebar_toggle,
)
from ._tooltip import tooltip, tooltip_toggle, update_tooltip
from ._valuebox import showcase_left_center, showcase_top_right, value_box

__all__ = (
    # Css
    "CssUnit",
    "as_css_unit",
    "as_css_padding",
    # Sidebar
    "DeprecatedPanelMain",
    "DeprecatedPanelSidebar",
    "Sidebar",
    "layout_sidebar",
    "panel_main",
    "panel_sidebar",
    "sidebar",
    "sidebar_toggle",
    # Page
    "page_sidebar",
    "page_fillable",
    "page_navbar",
    # Navs
    "navset_bar",
    "navset_tab_card",
    "navset_pill_card",
    # Card
    "CardItem",
    "ImgContainer",
    "TagCallable",
    "WrapperCallable",
    "card",
    "card_header",
    "card_title",
    "card_body",
    "card_image",
    "card_footer",
    # Layout
    "layout_column_wrap",
    # Tooltip
    "tooltip",
    "tooltip_toggle",
    "update_tooltip",
    # ValueBox
    "value_box",
    "showcase_left_center",
    "showcase_top_right",
    # Fill
    "FillingLayout",
    # "bind_fill_role",
    "as_fill_carrier",
    "as_fillable_container",
    "as_fill_item",
    "remove_all_fill",
    "is_fill_carrier",
    "is_fillable_container",
    "is_fill_item",
    # Output
    "output_image",
    "output_plot",
    "output_ui",
    # input_text_area
    "input_text_area",
    # Accordion
    "AccordionPanel",
    "accordion",
    "accordion_panel",
    "accordion_panel_set",
    "accordion_panel_open",
    "accordion_panel_close",
    "accordion_panel_insert",
    "accordion_panel_remove",
    "update_accordion_panel",
)



from __future__ import annotations

__all__ = ("output_data_frame",)

from htmltools import HTMLDependency, Tag

from ... import __version__
from ..._namespaces import resolve_id
from ...experimental.ui._fill import bind_fill_role


def data_frame_deps() -> HTMLDependency:
    return HTMLDependency(
        name="shiny-data-frame-output",
        version=__version__,
        source={
            "package": "shiny",
            "subdir": "www/shared/dataframe",
        },
        script={"src": "dataframe.js", "type": "module"},
    )


def output_data_frame(id: str) -> Tag:
    """
    Create a output container for a data frame.

    Parameters
    ----------
    id
        An input id.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    :func:`~shiny.render.data_frame`
    """
    return bind_fill_role(
        Tag(
            "shiny-data-frame",
            data_frame_deps(),
            id=resolve_id(id),
        ),
        item=True,
        container=True,
    )



# !!`shiny.experimental`!! This file is a direct copy to `shiny/_output.py`. Functions that were not altered were removed. Inserted code as been marked with `NEW`&`/NEW`
# Related: https://github.com/rstudio/shiny/pull/3715

from __future__ import annotations

from typing import Optional

from htmltools import Tag, TagAttrValue, TagFunction, css, div, tags

from ..._namespaces import resolve_id
from ...types import MISSING, MISSING_TYPE
from ...ui._plot_output_opts import (
    BrushOpts,
    ClickOpts,
    DblClickOpts,
    HoverOpts,
    brush_opts,
    click_opts,
    dblclick_opts,
    format_opt_names,
    hover_opts,
)
from ._fill import as_fill_item, as_fillable_container


# @add_example()
def output_plot(
    id: str,
    width: str = "100%",
    height: str = "400px",
    *,
    inline: bool = False,
    click: bool | ClickOpts = False,
    dblclick: bool | DblClickOpts = False,
    hover: bool | HoverOpts = False,
    brush: bool | BrushOpts = False,
    # NEW
    fill: bool | MISSING_TYPE = MISSING,
    # /NEW
) -> Tag:
    """
    Create a output container for a static plot.

    Place a :func:`~shiny.render.plot` result in the user interface. See
    :func:`~shiny.render.plot` for more details on what types of plots are supported.

    Parameters
    ----------
    id
        An input id.
    width
        The CSS width, e.g. '400px', or '100%'.
    height
        The CSS height, e.g. '100%' or '600px'.
    inline
        If ``True``, the result is displayed inline.
    click
        This can be a boolean or an object created by :func:`~shiny.ui.click_opts`. The
        default is `False`, but if you use `True` (or equivalently, `click_opts()`), the
        plot will send coordinates to the server whenever it is clicked, and the value
        will be accessible via `input.xx_click()`, where `xx` is replaced with the ID of
        this plot. The input value will be a dictionary with `x` and `y` elements
        indicating the mouse position.
    dblclick
        This is just like the `click` parameter, but for double-click events.
    hover
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.hover_opts`. The default is `False`, but if you use `True` (or
        equivalently, `hover_opts()`), the plot will send coordinates to the server
        whenever it is clicked, and the value will be accessible via `input.xx_hover()`,
        where `xx` is replaced with the ID of this plot. The input value will be a
        dictionary with `x` and `y` elements indicating the mouse position. To control
        the hover time or hover delay type, use :func:`~shiny.ui.hover_opts`.
    brush
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.brush_opts`. The default is `False`, but if you use `True` (or
        equivalently, `brush_opts()`), the plot will allow the user to "brush" in the
        plotting area, and will send information about the brushed area to the server,
        and the value will be accessible via `input.plot_brush()`. Brushing means that
        the user will be able to draw a rectangle in the plotting area and drag it
        around. The value will be a named list with `xmin`, `xmax`, `ymin`, and `ymax`
        elements indicating the brush area. To control the brush behavior, use
        :func:`~shiny.ui.brush_opts`. Multiple `output_image`/`output_plot` calls may
        share the same `id` value; brushing one image or plot will cause any other
        brushes with the same `id` to disappear.

    Returns
    -------
    :
        A UI element

    See Also
    -------
    * :func:`~shiny.render.plot`
    * :func:`~shiny.ui.output_image`
    """

    # NEW
    if isinstance(fill, MISSING_TYPE):
        fill = not inline
    # /NEW

    res = output_image(
        id=id,
        width=width,
        height=height,
        inline=inline,
        click=click,
        dblclick=dblclick,
        hover=hover,
        brush=brush,
        # NEW
        fill=fill,
        # /NEW
    )
    res.add_class("shiny-plot-output")
    return res


# @add_example()
def output_image(
    id: str,
    width: str = "100%",
    height: str = "400px",
    *,
    inline: bool = False,
    click: bool | ClickOpts = False,
    dblclick: bool | DblClickOpts = False,
    hover: bool | HoverOpts = False,
    brush: bool | BrushOpts = False,
    # NEW
    fill: bool = False,
    # /NEW
) -> Tag:
    """
    Create a output container for a static image.

    Parameters
    ----------
    id
        An input id.
    width
        The CSS width, e.g. '400px', or '100%'.
    height
        The CSS height, e.g. '100%' or '600px'.
    inline
        If ``True``, the result is displayed inline.
    click
        This can be a boolean or an object created by :func:`~shiny.ui.click_opts`. The
        default is `False`, but if you use `True` (or equivalently, `click_opts()`), the
        plot will send coordinates to the server whenever it is clicked, and the value
        will be accessible via `input.xx_click()`, where `xx` is replaced with the ID of
        this plot. The input value will be a dictionary with `x` and `y` elements
        indicating the mouse position.
    dblclick
        This is just like the `click` parameter, but for double-click events.
    hover
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.hover_opts`. The default is `False`, but if you use `True` (or
        equivalently, `hover_opts()`), the plot will send coordinates to the server
        whenever it is clicked, and the value will be accessible via `input.xx_hover()`,
        where `xx` is replaced with the ID of this plot. The input value will be a
        dictionary with `x` and `y` elements indicating the mouse position. To control
        the hover time or hover delay type, use :func:`~shiny.ui.hover_opts`.
    brush
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.brush_opts`. The default is `False`, but if you use `True` (or
        equivalently, `brush_opts()`), the plot will allow the user to "brush" in the
        plotting area, and will send information about the brushed area to the server,
        and the value will be accessible via `input.plot_brush()`. Brushing means that
        the user will be able to draw a rectangle in the plotting area and drag it
        around. The value will be a named list with `xmin`, `xmax`, `ymin`, and `ymax`
        elements indicating the brush area. To control the brush behavior, use
        :func:`~shiny.ui.brush_opts`. Multiple `output_image`/`output_plot` calls may
        share the same `id` value; brushing one image or plot will cause any other
        brushes with the same `id` to disappear.

    Returns
    -------
    :
        A UI element

    See Also
    -------
    ~shiny.render.image
    ~shiny.ui.output_plot
    """
    func = tags.span if inline else div
    style = None if inline else css(width=width, height=height)

    args: dict[str, str] = dict()

    id_resolved = resolve_id(id)

    if click is not False:
        if click is True:
            click = click_opts()
        click["id"] = id_resolved + "_click"
        args.update(**format_opt_names(click, "click"))

    if dblclick is not False:
        if dblclick is True:
            dblclick = dblclick_opts()
        dblclick["id"] = id_resolved + "_dblclick"
        args.update(**format_opt_names(dblclick, "dblclick"))

    if hover is not False:
        if hover is True:
            hover = hover_opts()
        hover["id"] = id_resolved + "_hover"
        args.update(**format_opt_names(hover, "hover"))

    if brush is not False:
        if brush is True:
            brush = brush_opts()
        brush["id"] = id_resolved + "_brush"
        args.update(**format_opt_names(brush, "brush"))

    container = func(
        id=id_resolved,
        class_="shiny-image-output",
        style=style,
        **args,
    )
    if fill:
        container = as_fill_item(container)

    return container


# @add_example()
def output_ui(
    id: str,
    inline: bool = False,
    container: Optional[TagFunction] = None,
    # NEW
    fill: bool = False,
    fillable: bool = False,
    # /NEW
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a output container for a UI (i.e., HTML) element.

    Parameters
    ----------
    id
        An input id.
    inline
        If ``True``, the result is displayed inline
    container
        A Callable that returns the output container.
    kwargs
        Attributes to be applied to the output container.

    Returns
    -------
    :
        A UI element

    See Also
    -------
    ~shiny.render.ui
    ~shiny.ui.output_text
    """

    if not container:
        container = tags.span if inline else tags.div
    res = container(
        {"class": "shiny-html-output"},
        id=resolve_id(id),
        **kwargs,
    )

    if fillable:
        res = as_fillable_container(res)
    if fill:
        res = as_fill_item(res)

    return res



from __future__ import annotations

import shiny.experimental as x
from shiny import App, Inputs, Outputs, Session, reactive, ui


def make_panel(letter: str) -> x.ui.AccordionPanel:
    return x.ui.accordion_panel(
        f"Section {letter}",
        f"Some narrative for section {letter}",
        value=f"sec_{letter}",
    )


items = [make_panel(letter) for letter in "ABCDE"]

app_ui = ui.page_fluid(
    ui.input_switch("update_panel", "Update Sections"),
    x.ui.accordion(*items, id="acc", multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    @reactive.event(input.update_panel)
    def _():
        txt = " (updated)" if input.update_panel() else ""
        for letter in "ABCDE":
            x.ui.update_accordion_panel(
                "acc",
                f"sec_{letter}",
                f"Some{txt} narrative for section {letter}",
                title=f"Section {letter}{txt}",
            )


app = App(app_ui, server)



from __future__ import annotations

from typing import Literal, Optional, Protocol, TypeVar, runtime_checkable

from htmltools import Tag, TagChild, Tagifiable, css

from ._css_unit import CssUnit, as_css_unit
from ._htmldeps import fill_dependency
from ._tag import tag_add_style, tag_prepend_class, tag_remove_class

"""
examples:
    - [ ] remove_all_fill
    - [ ] is_fill_carrier
    - [ ] is_fillable_container
    - [ ] is_fill_item
    - [ ] FillingLayout
"""


__all__ = (
    "bind_fill_role",
    "as_fill_carrier",
    "as_fillable_container",
    "as_fill_item",
    "remove_all_fill",
    "is_fill_carrier",
    "is_fillable_container",
    "is_fill_item",
    "FillingLayout",
)

TagFillingLayoutT = TypeVar("TagFillingLayoutT", bound="Tag | FillingLayout")
TagT = TypeVar("TagT", bound="Tag")


FILL_ITEM_CLASS = "html-fill-item"
FILL_CONTAINER_CLASS = "html-fill-container"

# We're currently not supporting the div(as_fillable_container()) API, in order to keep
# the function easier to understand. If we do implement something like that, we should
# use a different name, because the function is conceptually different, and it also is
# behaviorally different (it doesn't include the HTML dependency.)

# TODO-future-approach: bind_fill_role() should return None?
# From @wch:
# > For functions like this, which modify the original object, I think the Pythonic way
# > of doing things is to return None, to make it clearer that the object is modified in
# > place.
# From @schloerke:
# > It makes for a very clunky interface. Keeping as is for now.
# > Should we copy the tag before modifying it? (If we are not doing that elsewhere, then I am hesitant to start here.)
# > If it is not utilizing `nonlocal foo`, then it should be returned. Even if it is altered in-place


def _add_role(
    tag: TagT, *, condition: bool | None, class_: str, overwrite: bool = False
) -> TagT:
    if condition is None:
        return tag

    # Remove the class if it already exists and we're going to add it,
    # or if we're requiring it to be removed
    if (condition and tag.has_class(class_)) or overwrite:
        tag = tag_remove_class(tag, class_)

    if condition:
        tag = tag_prepend_class(tag, class_)
        tag.append(fill_dependency())
    return tag


def bind_fill_role(
    tag: TagT,
    *,
    item: Optional[bool] = None,
    container: Optional[bool] = None,
    overwrite: bool = False,
) -> TagT:
    """
    Allow tags to intelligently fill their container

    Create fill containers and items. If a fill item is a direct child of a fill
    container, and that container has an opinionated height, then the item is allowed to
    grow and shrink to its container's size.

    Parameters
    ----------
    tag
        a T object.
    item
        whether or not to treat `tag` as a fill item.
    container
        whether or not to treat `x` as a fill container. Note, this will set the CSS
        `display` property on the tag to `flex` which can change how its direct children
        are rendered. Thus, one should be careful not to mark a tag as a fill container
        when it needs to rely on other `display` behavior.
    overwrite
        whether or not to override previous filling layout calls (e.g., to remove the
        item/container role from a tag).

    Returns
    -------
    :
        The original :class:`~htmltools.Tag` object (`tag`) with additional attributes
        (and an :class:`~htmltools.HtmlDependency`).
    """
    tag = _add_role(
        tag,
        condition=item,
        class_=FILL_ITEM_CLASS,
        overwrite=overwrite,
    )
    tag = _add_role(
        tag,
        condition=container,
        class_=FILL_CONTAINER_CLASS,
        overwrite=overwrite,
    )
    return tag


###########################################


# Test and/or coerce fill behavior

# TODO-future; When `css_selector` can be implemented, the three parameters below should be added where appropriate
# @param class A character vector of class names to add to the tag.
# @param style A character vector of CSS properties to add to the tag.
# @param css_selector A character string containing a CSS selector for
#   targeting particular (inner) tag(s) of interest. For more details on what
#   selector(s) are supported, see [tagAppendAttributes()].


def as_fill_carrier(
    tag: TagFillingLayoutT,
    *,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    gap: Optional[CssUnit] = None,
    # class_: Optional[str] = None,
    # style: Optional[str] = None,
    # css_selector: Optional[str],
) -> TagFillingLayoutT:
    """
    Make a tag a fill carrier

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.
    min_height,max_height,gap
        Any valid CSS unit (e.g., `150`) to be applied to `tag`.

    Returns
    -------
    :
        The original :class:`~htmltools.Tag` object (`tag`) with additional attributes
        (and an :class:`~htmltools.HtmlDependency`).

    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_item`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.remove_all_fill`
    * :func:`~shiny.experimental.ui.is_fill_carrier`
    * :func:`~shiny.experimental.ui.is_fill_item`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """
    return _add_filling_attrs(
        tag,
        item=True,
        container=True,
        min_height=min_height,
        max_height=max_height,
        gap=gap,
    )


def as_fillable_container(
    tag: TagFillingLayoutT,
    *,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    gap: Optional[CssUnit] = None,
    # class_: Optional[str] = None,
    # style: Optional[str] = None,
    # css_selector: Optional[str] = None,
) -> TagFillingLayoutT:
    """
    Coerce a tag to be a fillable container

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.
    min_height,max_height,gap
        Any valid CSS unit (e.g., `150`) to be applied to `tag`.

    Returns
    -------
    :
        The original :class:`~htmltools.Tag` object (`tag`) with additional attributes
        (and an :class:`~htmltools.HtmlDependency`).

    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_carrier`
    * :func:`~shiny.experimental.ui.as_fill_item`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.remove_all_fill`
    * :func:`~shiny.experimental.ui.is_fill_carrier`
    * :func:`~shiny.experimental.ui.is_fill_item`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """
    return _add_filling_attrs(
        tag,
        container=True,
        min_height=min_height,
        max_height=max_height,
        gap=gap,
    )


def as_fill_item(
    tag: TagFillingLayoutT,
    *,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    # class_: Optional[str] = None,
    # style: Optional[str] = None,
    # css_selector: Optional[str] = None,
) -> TagFillingLayoutT:
    """
    Coerce a tag to a fill item

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.
    min_height,max_height
        Any valid CSS unit (e.g., `150`) to be applied to `tag`.

    Returns
    -------
    :
        The original :class:`~htmltools.Tag` object (`tag`) with additional attributes
        (and an :class:`~htmltools.HtmlDependency`).

    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_carrier`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.remove_all_fill`
    * :func:`~shiny.experimental.ui.is_fill_carrier`
    * :func:`~shiny.experimental.ui.is_fill_item`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """
    return _add_filling_attrs(
        tag,
        item=True,
        min_height=min_height,
        max_height=max_height,
    )


def remove_all_fill(tag: TagFillingLayoutT) -> TagFillingLayoutT:
    """
    Remove any filling layouts from a tag

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        The original :class:`~htmltools.Tag` object with filling layout attributes
        removed.


    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_carrier`
    * :func:`~shiny.experimental.ui.as_fill_item`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.is_fill_carrier`
    * :func:`~shiny.experimental.ui.is_fill_item`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """

    if isinstance(tag, FillingLayout):
        return tag.remove_all_fill()

    return bind_fill_role(
        tag,
        item=False,
        container=False,
        overwrite=True,
    )


def is_fill_carrier(tag: Tag | FillingLayout) -> bool:
    """
    Test a tag for being a fill carrier

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        Whether or not `tag` is a fill carrier.

    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_carrier`
    * :func:`~shiny.experimental.ui.as_fill_item`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.remove_all_fill`
    * :func:`~shiny.experimental.ui.is_fill_item`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """
    return is_fillable_container(tag) and is_fill_item(tag)


def is_fillable_container(tag: TagChild | FillingLayout) -> bool:
    """
    Test a tag for being a fillable container

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        Whether or not `tag` is a fillable container.


    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_carrier`
    * :func:`~shiny.experimental.ui.as_fill_item`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.remove_all_fill`
    * :func:`~shiny.experimental.ui.is_fill_item`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """
    # TODO-future; Handle widgets
    # # won't actually work until (htmltools#334) gets fixed
    # renders_to_tag_class(x, FILL_CONTAINER_CLASS, ".html-widget")

    return _is_fill_layout(tag, layout="fillable")


def is_fill_item(tag: TagChild | FillingLayout) -> bool:
    """
    Test a tag for being a fill item

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.experimental.ui.card`,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.layout_sidebar`) possess both `fillable` and `fill`
    arguments (to control their fill behavior). However, sometimes it's useful to add,
    remove, and/or test fillable/fill properties on arbitrary :class:`~htmltools.Tag`,
    which these functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        Whether or not `tag` is a fill item.

    See Also
    --------
    * :func:`~shiny.experimental.ui.as_fill_carrier`
    * :func:`~shiny.experimental.ui.as_fill_item`
    * :func:`~shiny.experimental.ui.as_fillable_container`
    * :func:`~shiny.experimental.ui.remove_all_fill`
    * :func:`~shiny.experimental.ui.is_fill_carrier`
    * :func:`~shiny.experimental.ui.is_fillable_container`
    """
    # TODO-future; Handle widgets
    # # won't actually work until (htmltools#334) gets fixed
    # renders_to_tag_class(x, FILL_ITEM_CLASS, ".html-widget")

    return _is_fill_layout(tag, layout="fill")


def _is_fill_layout(
    tag: TagChild | FillingLayout,
    layout: Literal["fill", "fillable"],
    # recurse: bool = True,
) -> bool:
    if not isinstance(tag, (Tag, Tagifiable, FillingLayout)):
        return False

    # tag: Tag | FillingLayout | Tagifiable

    if layout == "fill":
        if isinstance(tag, Tag):
            return tag.has_class(FILL_ITEM_CLASS)
        if isinstance(tag, FillingLayout):
            return tag.is_fill_item()

    elif layout == "fillable":
        if isinstance(tag, Tag):
            return tag.has_class(FILL_CONTAINER_CLASS)
        if isinstance(tag, FillingLayout):
            return tag.is_fillable_container()

    # tag: Tagifiable and not (Tag or FillingLayout)
    raise TypeError(
        f"`_is_fill_layout(tag=)` must be a `Tag` or implement the `FillingLayout` protocol methods. Received object of type: `{type(tag).__name__}`"
    )


T = TypeVar("T")


@runtime_checkable
class FillingLayout(Protocol):
    """
    Generic protocol for filling layouts objects
    """

    def add_class(
        self: T,
        class_: str,
        # *,
        # # Currently Unused
        # css_selector: Optional[str] = None,
        # Currently Unused
        **kwargs: object,
    ) -> T:
        """
        Generic method to handle adding a CSS `class` to an object

        Parameters
        ----------
        class_
            A character vector of class names to add to the tag.
        **kwargs
            Possible future arguments

        Returns
        -------
        :
            The updated object.
        """
        ...

    def add_style(
        self: T,
        style: str,
        # *,
        # # Currently Unused
        # css_selector: Optional[str] = None,
        # Currently Unused
        **kwargs: object,
    ) -> T:
        """
        Generic method to handle adding a CSS `style` to an object

        Parameters
        ----------
        style
            A character vector of CSS properties to add to the tag.
        **kwargs
            Possible future arguments

        Returns
        -------
        :
            The updated object.
        """
        ...

    def is_fill_item(self) -> bool:
        """
        Generic method to handle testing if an object is a fill item

        Returns
        -------
        :
            Whether or not the object is a fill item
        """
        ...

    def is_fillable_container(self) -> bool:
        """
        Generic method to handle testing if an object is a fillable container

        Returns
        -------
        :
            Whether or not the object is a fillable container
        """
        ...

    def as_fill_item(
        self: T,
    ) -> T:
        """
        Generic method to handle coercing an object to a fill item

        Returns
        -------
        :
            The updated object.
        """
        ...

    def as_fillable_container(
        self: T,
    ) -> T:
        """
        Generic method to handle coercing an object to a fillable container

        Returns
        -------
        :
            The updated object.
        """
        ...

    def remove_all_fill(
        self: T,
    ) -> T:
        """
        Generic method to handle removing all fill properties from an object

        Returns
        -------
        :
            The updated object.
        """
        ...


def _style_units_to_str(**kwargs: CssUnit | None) -> str | None:
    style_items: dict[str, CssUnit] = {}
    for k, v in kwargs.items():
        if v is not None:
            style_items[k] = as_css_unit(v)

    return css(**style_items)


def _add_filling_attrs(
    tag: TagFillingLayoutT,
    item: Optional[bool] = None,
    container: Optional[bool] = None,
    **kwargs: CssUnit | None,
) -> TagFillingLayoutT:
    new_style = _style_units_to_str(**kwargs)

    if isinstance(tag, FillingLayout):
        if new_style:
            tag.add_style(new_style)
        if item:
            tag.as_fill_item()
        if container:
            tag.as_fillable_container()
        return tag

    # Tag
    tag = tag_add_style(tag, new_style)
    return bind_fill_role(tag, item=item, container=container)



from __future__ import annotations

import htmltools

import shiny.experimental as x
from shiny import App, ui


def outer_inner() -> tuple[htmltools.Tag, htmltools.Tag]:
    inner = ui.div(
        id="inner",
        style=htmltools.css(
            height="200px",
            border="3px blue solid",
        ),
    )
    outer = ui.div(
        inner,
        id="outer",
        style=htmltools.css(
            height="300px",
            border="3px red solid",
        ),
    )
    return outer, inner


outer0, inner0 = outer_inner()
outer1, inner1 = outer_inner()
outer2, inner2 = outer_inner()

x.ui.as_fillable_container(outer2)

x.ui.as_fillable_container(outer2)
x.ui.as_fill_item(inner2)


app_ui = ui.page_fluid(
    ui.markdown(
        """\
        # `as_fill_container()`

        For an item to fill its parent element,
        * the item must have `as_fill_item()` be called on it
        * the parent container must have `as_fillable_container()` called on it

        Iff both methods are called, the inner child will naturally expand into its parent container.
        """
    ),
    ui.row(
        ui.column(4, ui.h5("Default behavior")),
        ui.column(4, ui.h5(ui.markdown("`as_fill_container(red)`"))),
        ui.column(
            4,
            ui.h5(ui.markdown("`as_fill_item(blue)` + `as_fillable_container(red)`")),
        ),
    ),
    ui.row(
        ui.column(4, ui.div(outer0)),
        ui.column(4, ui.div(outer1)),
        ui.column(4, x.ui.as_fill_carrier(ui.span(outer2))),
    ),
)


app = App(app_ui, server=None)



from __future__ import annotations

import json
from typing import Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, TagList, div

from ... import Session
from ..._utils import drop_none
from ...session import require_active_session

# from ._color import get_color_contrast
from ._utils import consolidate_attrs
from ._web_component import web_component


def tooltip(
    trigger: TagChild,
    *args: TagChild | TagAttrs,
    id: Optional[str] = None,
    placement: Literal["auto", "top", "right", "bottom", "left"] = "auto",
    options: Optional[dict[str, object]] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Add a tooltip to a UI element

    Display additional information when focusing (or hovering over) a UI element.

    Parameters
    ----------
    trigger
        A UI element (i.e., :class:`~htmltools.Tag`) to serve as the tooltips trigger.
        It's good practice for this element to be a keyboard-focusable and interactive
        element (e.g., :func:`~shiny.ui.input_action_button`,
        :func:`~shiny.ui.input_action_link`, etc.) so that the tooltip is accessible to
        keyboard and assistive technology users.
    *args
        Contents to the tooltip's body. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    id
        A character string. Required to re-actively respond to the visibility of the
        tooltip (via the `input[id]` value) and/or update the visibility/contents of the
        tooltip.
    placement
        The placement of the tooltip relative to its trigger.
    options
        A list of additional [Bootstrap
        options](https://getbootstrap.com/docs/5.2/components/tooltips/#options).

    Details
    -------

    If `trigger` yields multiple HTML elements (e.g., a :class:`~htmltools.TagList` or
    complex [`shinywidgets`](https://github.com/rstudio/py-shinywidgets) object), the
    last HTML element is used as the trigger. If the `trigger` should contain all of
    those elements, wrap the object in a :func:`~htmltools.div` or :func:`~htmltools.span`.

    See Also
    --------

    * [Bootstrap tooltips documentation](https://getbootstrap.com/docs/5.2/components/tooltips/)
    """
    attrs, children = consolidate_attrs(*args, **kwargs)

    if len(children) == 0:
        raise RuntimeError("At least one value must be provided to `*args: TagChild`")

    res = web_component(
        "bslib-tooltip",
        {
            "id": id,
            "placement": placement,
            "options": json.dumps(options) if options else None,
        },
        attrs,
        # Use display:none instead of <template> since shiny.js
        # doesn't bind to the contents of the latter
        div(*children, {"style": "display:none;"}),
        trigger,
    )

    return res


def _session_on_flush_send_msg(
    id: str, session: Session | None, msg: dict[str, object]
) -> None:
    session = require_active_session(session)
    session.on_flush(lambda: session.send_input_message(id, msg), once=True)


def tooltip_toggle(
    id: str, show: Optional[bool] = None, session: Optional[Session] = None
) -> None:
    """
    Programmatically show/hide a tooltip

    Parameters
    ----------
    id
        A character string that matches an existing tooltip id.
    show
        Whether to show (`True`) or hide (`False`) the tooltip. The default (`None`)
        will show if currently hidden and hide if currently shown. Note that a tooltip
        will not be shown if the trigger is not visible (e.g., it's hidden behind a
        tab).
    session
        A Shiny session object (the default should almost always be used).
    """
    _session_on_flush_send_msg(
        id,
        session,
        {
            "method": "toggle",
            "value": _normalize_show_value(show),
        },
    )


# @describeIn tooltip Update the contents of a tooltip.
# @export
def update_tooltip(id: str, *args: TagChild, session: Optional[Session] = None) -> None:
    _session_on_flush_send_msg(
        id,
        session,
        drop_none(
            {
                "method": "update",
                "title": require_active_session(session)._process_ui(TagList(*args))
                if len(args) > 0
                else None,
            }
        ),
    )


def _normalize_show_value(show: bool | None) -> Literal["toggle", "show", "hide"]:
    if show is None:
        return "toggle"
    return "show" if show else "hide"



from __future__ import annotations

import shiny.experimental as x
from shiny import App, Inputs, Outputs, Session, reactive, ui

items = [
    x.ui.accordion_panel(f"Section {letter}", f"Some narrative for section {letter}")
    for letter in "ABCDE"
]

# # First shown by default
# x.ui.accordion(*items)

# # Nothing shown by default
# x.ui.accordion(*items, open=False)
# # Everything shown by default
# x.ui.accordion(*items, open=True)

# # Show particular sections
# x.ui.accordion(*items, open="Section B")
# x.ui.accordion(*items, open=["Section A", "Section B"])

app_ui = ui.page_fluid(
    # Provide an id to create a shiny input binding
    x.ui.accordion(*items, id="acc"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    def _():
        print(input.acc())


app = App(app_ui, server)



from __future__ import annotations

from typing import Optional, overload

from htmltools import (
    HTMLDependency,
    Tag,
    TagChild,
    TagList,
    TagNode,
    head_content,
    tags,
)

from ..types import MISSING, MISSING_TYPE


def shiny_input_label(id: str, label: TagChild = None) -> Tag:
    cls = "control-label" + ("" if label else " shiny-label-null")
    return tags.label(label, class_=cls, id=id + "-label", for_=id)


@overload
def get_window_title(
    title: None,
    window_title: MISSING_TYPE,
) -> None:
    ...


@overload
def get_window_title(
    title: None,
    window_title: str,
) -> HTMLDependency:
    ...


@overload
def get_window_title(
    title: str | Tag | TagList,
    window_title: str | MISSING_TYPE,
) -> HTMLDependency:
    ...


def get_window_title(
    title: Optional[str | Tag | TagList],
    window_title: str | MISSING_TYPE = MISSING,
) -> Optional[HTMLDependency]:
    if title is not None and isinstance(window_title, MISSING_TYPE):
        window_title = _find_child_strings(title)

    if isinstance(window_title, MISSING_TYPE):
        return None
    else:
        return head_content(tags.title(window_title))


def _find_child_strings(x: TagList | TagNode) -> str:
    if isinstance(x, Tag) and x.name not in ("script", "style"):
        x = x.children
    if isinstance(x, TagList):
        strings = [_find_child_strings(y) for y in x]
        return " ".join(filter(lambda x: x != "", strings))
    if isinstance(x, str):
        return x
    return ""



# Needed for types imported only during TYPE_CHECKING with Python 3.7 - 3.9
# See https://www.python.org/dev/peps/pep-0655/#usage-in-python-3-11
from __future__ import annotations

import re
from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import Pattern, Union


class ResolvedId(str):
    def __call__(self, id: Id) -> ResolvedId:
        if isinstance(id, ResolvedId):
            return id

        validate_id(id)

        if self == "":
            return ResolvedId(id)
        else:
            return ResolvedId(self + "-" + id)


Root: ResolvedId = ResolvedId("")


Id = Union[str, ResolvedId]


def current_namespace() -> ResolvedId:
    return _current_namespace.get()


def resolve_id(id: Id) -> ResolvedId:
    """
    Resolve an ID, possibly with a module namespace.

    Parameters
    ----------
    Args
        id: An ID.

    Returns
        An ID (if in a module, this will contain a namespace prefix).
    """
    curr_ns = _current_namespace.get()
    return curr_ns(id)


# \w is a large set for unicode patterns, that's fine; we mostly want to avoid some
# special characters like space, comma, period, and especially dash
re_valid_id: Pattern[str] = re.compile("^\\.?\\w+$")


def validate_id(id: str):
    if not re_valid_id.match(id):
        raise ValueError(
            f"The string '{id}' is not a valid id; only letters, numbers, and "
            "underscore are permitted"
        )


_current_namespace: ContextVar[ResolvedId] = ContextVar(
    "current_namespace", default=Root
)


@contextmanager
def namespace_context(id: Id | None):
    namespace = resolve_id(id) if id else Root
    token: Token[ResolvedId] = _current_namespace.set(namespace)
    try:
        yield
    finally:
        _current_namespace.reset(token)



from __future__ import annotations

import shiny.experimental as x
from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = x.ui.page_sidebar(
    x.ui.sidebar("Sidebar content", id="sidebar"),
    ui.input_action_button(
        "sidebar_toggle",
        label="Toggle sidebar",
        width="fit-content",
    ),
    ui.output_text_verbatim("state"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    @reactive.event(input.sidebar_toggle)
    def _():
        x.ui.sidebar_toggle("sidebar")

    @output
    @render.text
    def state():
        return f"input.sidebar(): {input.sidebar()}"


app = App(app_ui, server=server)



from __future__ import annotations

from pathlib import PurePath

from htmltools import HTMLDependency

from ..._versions import bslib as bslib_version
from ..._versions import htmltools as htmltools_version

_x_www = PurePath(__file__).parent.parent.parent / "www" / "shared" / "_x"
_x_sidebar_path = str(_x_www / "bslib" / "components" / "sidebar")
_x_nav_spacer_path = str(_x_www / "bslib" / "components" / "nav_spacer")
_x_fill_path = str(_x_www / "htmltools" / "fill")


def fill_dependency() -> HTMLDependency:
    return HTMLDependency(
        "htmltools-fill",
        htmltools_version,
        source={
            "package": "shiny",
            "subdir": _x_fill_path,
        },
        stylesheet={"href": "fill.css"},
    )


def sidebar_dependency() -> HTMLDependency:
    return HTMLDependency(
        "bslib-sidebar",
        bslib_version,
        source={
            "package": "shiny",
            "subdir": _x_sidebar_path,
        },
        script={"src": "sidebar.min.js"},
        all_files=True,
    )


def nav_spacer_dependency() -> HTMLDependency:
    return HTMLDependency(
        "bslib-nav-spacer",
        bslib_version,
        source={
            "package": "shiny",
            "subdir": _x_nav_spacer_path,
        },
        stylesheet={"href": "nav_spacer.css"},
    )



from __future__ import annotations

from typing import Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, css, div

from ._css_unit import CssUnit, as_css_unit
from ._fill import as_fill_item, as_fillable_container
from ._htmldeps import grid_dependency
from ._utils import consolidate_attrs, is_01_scalar


def layout_column_wrap(
    width: Optional[CssUnit],
    *args: TagChild | TagAttrs,
    fixed_width: bool = False,
    heights_equal: Literal["all", "row"] = "all",
    fill: bool = True,
    fillable: bool = True,
    height: Optional[CssUnit] = None,
    height_mobile: Optional[CssUnit] = None,
    gap: Optional[CssUnit] = None,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    # For
    # more explanation and illustrative examples, see
    # [here](https://rstudio.github.io/bslib/articles/cards.html#multiple-cards)
    """
    A grid-like, column-first layout

    Wraps a 1d sequence of UI elements into a 2d grid. The number of columns (and rows)
    in the grid dependent on the column `width` as well as the size of the display.

    Parameters
    ----------
    width
        The desired width of each card. It can be a (unit-less) number between 0 and 1
        and should be specified as `1/num`, where `num` represents the number of desired
        columns. It can be a CSS length unit representing either the minimum (when
        `fixed_width=False`) or fixed width (`fixed_width=True`). It can also be `None`,
        which allows power users to set the `grid-template-columns` CSS property
        manually, either via a `style` attribute or a CSS stylesheet.
    *args
        Unnamed arguments should be UI elements (e.g.,
        :func:`~shiny.experimental.ui.card`). Named arguments become attributes on the
        containing :class:`~htmltools.Tag` element.
    fixed_width
        Whether or not to interpret the `width` as a minimum (`fixed_width=False`) or
        fixed (`fixed_width=True`) width when it is a CSS length unit.
    heights_equal
        If `"all"` (the default), every card in every row of the grid will have the same
        height. If `"row"`, then every card in _each_ row of the grid will have the same
        height, but heights may vary between rows.
    fill
        Whether or not to allow the layout to grow/shrink to fit a fillable container
        with an opinionated height (e.g., :func:`~shiny.experimental.ui.page_fillable`).
    fillable
        Whether or not each element is wrapped in a fillable container.
    height
        Any valid CSS unit to use for the height.
    height_mobile
        Any valid CSS unit to use for the height when on mobile devices (or narrow
        windows).
    gap
        Any valid CSS unit to use for the gap between columns.
    class_
        A CSS class to apply to the containing element.
    **kwargs
        Additional attributes to apply to the containing element.

    Returns
    -------
    :
        A :class:`~htmltools.Tag` element.
    """
    attrs, children = consolidate_attrs(*args, class_=class_, **kwargs)

    colspec: str | None = None
    if width is not None:
        if is_01_scalar(width) and width > 0.0:
            num_cols = 1.0 / width
            if not num_cols.is_integer():
                raise ValueError(
                    "Could not interpret `layout_column_wrap(width=)` argument"
                )
            colspec = " ".join(["1fr" for _ in range(int(num_cols))])
        else:
            width_css_unit = as_css_unit(width)
            if fixed_width:
                colspec = f"repeat(auto-fit, minmax({width_css_unit}, 1fr))"
            else:
                colspec = f"repeat(auto-fit, minmax(0, {width_css_unit}))"

    # Use a new dict so that we don't mutate the original `children` dict
    upgraded_children: list[TagChild] = []
    for child_value in children:
        child = div({"class": "bslib-gap-spacing"}, child_value)
        if fillable:
            child = as_fillable_container(child)
        upgraded_children.append(child)

    tag_style_css = {
        "grid-template-columns": colspec,
        "grid-auto-rows": "1fr" if (heights_equal == "all") else None,
        # Always provide the `height:auto` default so that the CSS variable
        # doesn't get inherited in a scenario like
        # layout_column_wrap(height=200, layout, layout_column_wrap(...))
        "--bslib-grid-height": as_css_unit("auto" if height is None else height),
        "--bslib-grid-height-mobile": as_css_unit(
            "auto" if height_mobile is None else height_mobile
        ),
        "gap": as_css_unit(gap),
    }

    tag = div(
        {
            "class": "bslib-grid",
            "style": css(**tag_style_css),
        },
        attrs,
        *upgraded_children,
        grid_dependency(),
    )
    if fill:
        tag = as_fill_item(tag)

    return tag



from __future__ import annotations

from typing import TypeVar

from htmltools import Tag

TagT = TypeVar("TagT", bound="Tag")

# Tag.add_class(x: str) -> Self[Tag]:
#     cls = self.attrs.get("class")
#     if cls:
#         x = cls + " " + x
#     self.attrs["class"] = x
#     return self


# Do not export
# "Prepend" version of `tag.add_class(class_)`
def tag_prepend_class(tag: TagT, *class_: str | None) -> TagT:
    classes = (
        *class_,
        tag.attrs.get("class"),
    )
    classes = [c for c in classes if c is not None]
    if len(classes) == 0:
        return tag
    tag.attrs["class"] = " ".join(classes)
    return tag



from __future__ import annotations

from typing import Callable, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, css, div, tags

from ._card import CardItem, card, card_body
from ._css_unit import CssUnit, as_css_unit, as_width_unit
from ._fill import as_fill_carrier
from ._htmldeps import value_box_dependency
from ._layout import layout_column_wrap
from ._utils import consolidate_attrs, is_01_scalar

__all__ = (
    "value_box",
    "showcase_left_center",
    "showcase_top_right",
)


# TODO-maindocs; @add_example()
def value_box(
    title: TagChild,
    value: TagChild,
    *args: TagChild | TagAttrs,
    showcase: Optional[TagChild] = None,
    showcase_layout: Callable[[TagChild, Tag], CardItem] | None = None,
    full_screen: bool = False,
    theme_color: Optional[str] = "primary",
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Value box

    An opinionated (:func:`~shiny.experimental.ui.card`-powered) box, designed for
    displaying a `value` and `title`. Optionally, a `showcase` can provide for context
    for what the `value` represents (for example, it could hold an icon, or even a
    :func:`~shiny.ui.output_plot`).

    Parameters
    ----------
    title,value
        A string, number, or :class:`~htmltools.Tag` child to display as
        the title or value of the value box. The `title` appears above the `value`.
    *args
        Unnamed arguments may be any :class:`~htmltools.Tag` children to display below
        `value`. Named arguments are passed to :func:`~shiny.experimental.ui.card` as
        element attributes.
    showcase
        A :class:`~htmltools.Tag` child to showcase (e.g., an icon, a
        :func:`~shiny.ui.output_plot`, etc).
    showcase_layout
        Either :func:`~shiny.experimental.ui.showcase_left_center` or
        :func:`~shiny.experimental.ui.showcase_top_right`.
    theme_color
        A theme color to use for the background color. Should match a name in the
        Bootstrap Sass variable `$theme-colors` (e.g., `"secondary"`, `"success"`,
        `"danger"`, etc).
    height,max_height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`).
    fill
        Whether to allow the value box to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.experimental.ui.page_fillable`).
    class_
        Utility classes for customizing the appearance of the summary card. Use `bg-*`
        and `text-*` classes (e.g, `"bg-danger"` and `"text-light"`) to customize the
        background/foreground colors.
    **kwargs
        Additional attributes to pass to :func:`~shiny.experimental.ui.card`.

    Returns
    -------
    :
        A :func:`~shiny.experimental.ui.card`

    See Also
    --------
    * :func:`~shiny.experimental.ui.card`
    """
    attrs, children = consolidate_attrs(
        # Must be before `attrs` so that `class_` is applied before any `attrs` values
        {"class": "bslib-value-box border-0"},
        {"class": f"bg-{theme_color}"} if theme_color else None,
        *args,
        class_=class_,
        **kwargs,
    )

    if showcase_layout is None:
        showcase_layout = showcase_left_center()
    if isinstance(title, (str, int, float)):
        title = tags.p(str(title), class_="h6 mb-1")
    if isinstance(title, (str, int, float)):
        value = tags.p(str(value), class_="h2 mb-2")

    contents = div(
        title,
        value,
        *children,
        class_="value-box-area",
    )
    contents = as_fill_carrier(contents)

    if showcase is not None:
        contents = showcase_layout(showcase, contents)

    return card(
        contents,
        attrs,
        value_box_dependency(),
        full_screen=full_screen,
        height=height,
        max_height=max_height,
        fill=fill,
    )


# TODO-maindocs; @add_example()
def showcase_left_center(
    width: CssUnit = "30%",
    max_height: CssUnit = "100px",
    max_height_full_screen: CssUnit = "67%",
) -> Callable[[TagChild | TagAttrs, Tag], CardItem]:
    """
    Left center showcase for a value box

    Gives the showcase a width and centers it vertically.

    Parameters
    ----------
    width
        A proportion (i.e., a number between 0 and 1) of available width to allocate to
        the showcase. Or, A vector of length 2 valid CSS unit defining the width of each
        column (for `showcase_left_center()` the 1st unit defines the showcase width and
        for `showcase_top_right` the 2nd unit defines the showcase width). Note that any
        units supported by the CSS grid `grid-template-columns` property may be used
        (e.g., `fr` units).
    max_height,max_height_full_screen
        A proportion (i.e., a number between 0 and 1) or any valid CSS unit defining the
        showcase max_height.

    Returns
    -------
    :
        A function that takes a showcase and contents and returns a :func:`~shiny.experimental.ui.card_body`
    """
    return _showcase_layout(
        width=width,
        max_height=max_height,
        max_height_full_screen=max_height_full_screen,
        top_right=False,
    )


# TODO-maindocs; @add_example()
def showcase_top_right(
    width: CssUnit = "30%",
    max_height: CssUnit = "75px",
    max_height_full_screen: CssUnit = "67%",
) -> Callable[[TagChild | TagAttrs, Tag], CardItem]:
    """
    Top right showcase for a value box

    Gives the showcase a width and in the top right corner.

    Parameters
    ----------
    width
        A proportion (i.e., a number between 0 and 1) of available width to allocate to
        the showcase. Or, A vector of length 2 valid CSS unit defining the width of each
        column (for `showcase_left_center()` the 1st unit defines the showcase width and
        for `showcase_top_right` the 2nd unit defines the showcase width). Note that any
        units supported by the CSS grid `grid-template-columns` property may be used
        (e.g., `fr` units).
    max_height,max_height_full_screen
        A proportion (i.e., a number between 0 and 1) or any valid CSS unit defining the
        showcase max_height.

    Returns
    -------
    :
        A function that takes a showcase and contents and returns a :func:`~shiny.experimental.ui.card_body`
    """

    if is_01_scalar(width):
        width = 1 - width
    return _showcase_layout(
        width=width,
        max_height=max_height,
        max_height_full_screen=max_height_full_screen,
        top_right=True,
    )


def _showcase_layout(
    width: CssUnit,
    max_height: CssUnit,
    max_height_full_screen: CssUnit,
    top_right: bool,
) -> Callable[[TagChild | TagAttrs, Tag], CardItem]:
    # Do not "magically" turn `0.3` into `"30%"` as it is not clear to the user when it happens
    width_css_unit = as_width_unit(width)
    max_height_css_unit = as_css_unit(max_height)
    max_height_full_screen_css_unit = as_css_unit(max_height_full_screen)

    def _layout(showcase: TagChild | TagAttrs, contents: Tag) -> CardItem:
        css_args = {
            "--bslib-value-box-max-height": max_height_css_unit,
            "--bslib-value-box-max-height-full-screen": max_height_full_screen_css_unit,
        }
        showcase_container = div(
            showcase,
            {"class": "value-box-showcase overflow-hidden"},
            {"class": "showcase-top-right"} if top_right else None,
            style=css(**css_args),
        )
        showcase_container = as_fill_carrier(showcase_container)

        if not top_right:
            contents.add_class("border-start")

        items = [showcase_container, contents]
        width_fs = ["1fr", "auto"]
        if top_right:
            items = reversed(items)
            width_fs = reversed(width_fs)

        width_fs = " ".join(width_fs)

        layout_css_args = {
            "--bslib-value-box-widths": width_css_unit,
            "--bslib-value-box-widths-full-screen": width_fs,
        }
        return card_body(
            layout_column_wrap(
                None,  # width
                *items,
                style=css(**layout_css_args),
                gap=0,
                heights_equal="row",
                class_="value-box-grid",
            ),
            style=css(padding=0),
        )

    return _layout



from __future__ import annotations

__all__ = (
    "navset_bar",
    "navset_tab_card",
    "navset_pill_card",
)

import copy
from typing import Any, Literal, Optional, Sequence, cast

from htmltools import MetadataNode, Tag, TagChild, TagList, css, div, tags

from ..._namespaces import resolve_id
from ..._utils import private_random_int
from ...types import NavSetArg
from ...ui._html_dependencies import bootstrap_deps
from ._card import CardItem, card, card_body, card_footer, card_header
from ._css_unit import CssUnit, as_css_padding, as_css_unit
from ._fill import as_fill_carrier
from ._sidebar import Sidebar, layout_sidebar
from ._tag import tag_add_style


# -----------------------------------------------------------------------------
# Navigation items
# -----------------------------------------------------------------------------
class Nav:
    nav: Tag
    content: Optional[Tag]

    def __init__(self, nav: Tag, content: Optional[Tag] = None) -> None:
        self.nav = nav
        # nav_control()/nav_spacer() have None as their content
        self.content = content

    def resolve(
        self, selected: Optional[str], context: dict[str, Any]
    ) -> tuple[TagChild, TagChild]:
        # Nothing to do for nav_control()/nav_spacer()
        if self.content is None:
            return self.nav, None

        # At least currently, in the case where both nav and content are tags
        # (i.e., nav()), the nav always has a child <a> tag...I'm not sure if
        # there's a way to statically type this
        nav = copy.deepcopy(self.nav)
        a_tag = cast(Tag, nav.children[0])
        if context.get("is_menu", False):
            a_tag.add_class("dropdown-item")
        else:
            a_tag.add_class("nav-link")
            nav.add_class("nav-item")

        # Hyperlink the nav to the content
        content = copy.copy(self.content)
        if "tabsetid" in context and "index" in context:
            id = f"tab-{context['tabsetid']}-{context['index']}"
            content.attrs["id"] = id
            a_tag.attrs["href"] = f"#{id}"

        # Mark the nav/content as active if it should be
        if isinstance(selected, str) and selected == self.get_value():
            content.add_class("active")
            a_tag.add_class("active")

        nav.children[0] = a_tag

        return nav, content

    def get_value(self) -> Optional[str]:
        if self.content is None:
            return None
        a_tag = cast(Tag, self.nav.children[0])
        return a_tag.attrs.get("data-value", None)

    def tagify(self) -> None:
        raise NotImplementedError(
            "nav() items must appear within navset_*() container."
        )


class NavSet:
    args: tuple[NavSetArg | MetadataNode]
    ul_class: str
    id: Optional[str]
    selected: Optional[str]
    header: TagChild
    footer: TagChild

    def __init__(
        self,
        *args: NavSetArg | MetadataNode,
        ul_class: str,
        id: Optional[str],
        selected: Optional[str],
        header: TagChild = None,
        footer: TagChild = None,
    ) -> None:
        self.args = args
        self.ul_class = ul_class
        self.id = id
        self.selected = selected
        self.header = header
        self.footer = footer

    def tagify(self) -> TagList | Tag:
        id = self.id
        ul_class = self.ul_class
        if id is not None:
            ul_class += " shiny-tab-input"

        nav, content = _render_navset(
            *self.args, ul_class=ul_class, id=id, selected=self.selected, context={}
        )
        return self.layout(nav, content).tagify()

    # Types must match output of `_render_navset() -> Tuple[Tag, Tag]`
    def layout(self, nav: Tag, content: Tag) -> TagList | Tag:
        return TagList(nav, self.header, content, self.footer)


# -----------------------------------------------------------------------------
# Navigation containers
# -----------------------------------------------------------------------------


class NavSetCard(NavSet):
    placement: Literal["above", "below"]
    sidebar: Optional[Sidebar]

    def __init__(
        self,
        *args: NavSetArg,
        ul_class: str,
        id: Optional[str],
        selected: Optional[str],
        sidebar: Optional[Sidebar] = None,
        header: TagChild = None,
        footer: TagChild = None,
        placement: Literal["above", "below"] = "above",
    ) -> None:
        super().__init__(
            *args,
            ul_class=ul_class,
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        )
        self.sidebar = sidebar
        self.placement = placement

    def layout(self, nav: Tag, content: Tag) -> Tag:
        # navs = [child for child in content.children if isinstance(child, Nav)]
        # not_navs = [child for child in content.children if child not in navs]
        content_val: Tag | CardItem = content

        if self.sidebar:
            content_val = navset_card_body(content, sidebar=self.sidebar)

        if self.placement == "below":
            # TODO-carson; have carson double check this change
            return card(
                card_header(self.header) if self.header else None,
                content_val,
                card_body(self.footer, fillable=False, fill=False)
                if self.footer
                else None,
                card_footer(nav),
            )
        else:
            # TODO-carson; have carson double check this change
            return card(
                card_header(nav),
                card_body(self.header, fill=False, fillable=False)
                if self.header
                else None,
                content_val,
                card_footer(self.footer) if self.footer else None,
            )


def navset_card_body(content: Tag, sidebar: Optional[Sidebar] = None) -> CardItem:
    content = _make_tabs_fillable(content, fillable=True, gap=0, padding=0)
    if sidebar:
        return layout_sidebar(
            sidebar,
            content,
            fillable=True,
            border=False,
        )
    else:
        return CardItem(content)


def navset_tab_card(
    *args: NavSetArg,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> NavSetCard:
    """
    Render nav items as a tabset inside a card container.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.

    See Also
    -------
    * ~shiny.ui.nav
    * ~shiny.ui.nav_menu
    * ~shiny.ui.nav_control
    * ~shiny.ui.nav_spacer
    * ~shiny.experimental.ui.navset_bar
    * ~shiny.ui.navset_tab
    * ~shiny.ui.navset_pill
    * ~shiny.experimental.ui.navset_pill_card
    * ~shiny.ui.navset_hidden

    Example
    -------
    See :func:`~shiny.ui.nav`
    """

    return NavSetCard(
        *args,
        ul_class="nav nav-tabs card-header-tabs",
        id=resolve_id(id) if id else None,
        selected=selected,
        sidebar=sidebar,
        header=header,
        footer=footer,
        placement="above",
    )


def navset_pill_card(
    *args: NavSetArg,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
    placement: Literal["above", "below"] = "above",
) -> NavSetCard:
    """
    Render nav items as a pillset inside a card container.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    placement
        Placement of the nav items relative to the content.

    See Also
    -------
    * ~shiny.ui.nav
    * ~shiny.ui.nav_menu
    * ~shiny.ui.nav_control
    * ~shiny.ui.nav_spacer
    * ~shiny.experimental.ui.navset_bar
    * ~shiny.ui.navset_tab
    * ~shiny.ui.navset_pill
    * ~shiny.experimental.ui.navset_tab_card
    * ~shiny.ui.navset_hidden

    Example
    -------
    See :func:`~shiny.ui.nav`
    """

    return NavSetCard(
        *args,
        ul_class="nav nav-pills card-header-pills",
        id=resolve_id(id) if id else None,
        selected=selected,
        sidebar=sidebar,
        header=header,
        footer=footer,
        placement=placement,
    )


class NavSetBar(NavSet):
    title: TagChild
    sidebar: Optional[Sidebar]
    fillable: bool | list[str]
    gap: Optional[CssUnit]
    padding: Optional[CssUnit | list[CssUnit]]
    position: Literal["static-top", "fixed-top", "fixed-bottom", "sticky-top"]
    bg: Optional[str]
    inverse: bool
    collapsible: bool
    fluid: bool

    def __init__(
        self,
        *args: NavSetArg | MetadataNode,
        ul_class: str,
        title: TagChild,
        id: Optional[str],
        selected: Optional[str],
        sidebar: Optional[Sidebar] = None,
        fillable: bool | list[str] = False,
        gap: Optional[CssUnit],
        padding: Optional[CssUnit | list[CssUnit]],
        position: Literal[
            "static-top", "fixed-top", "fixed-bottom", "sticky-top"
        ] = "static-top",
        header: TagChild = None,
        footer: TagChild = None,
        bg: Optional[str] = None,
        # TODO-bslib: default to 'auto', like we have in R (parse color via webcolors?)
        inverse: bool = False,
        collapsible: bool = True,
        fluid: bool = True,
    ) -> None:
        super().__init__(
            *args,
            ul_class=ul_class,
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        )
        self.title = title
        self.sidebar = sidebar
        self.fillable = fillable
        self.gap = gap
        self.padding = padding

        self.position = position
        self.bg = bg
        self.inverse = inverse
        self.collapsible = collapsible
        self.fluid = fluid

    def layout(self, nav: Tag, content: Tag) -> TagList:
        nav_container = div(
            {"class": "container-fluid" if self.fluid else "container"},
            tags.a({"class": "navbar-brand", "href": "#"}, self.title),
        )
        if self.collapsible:
            collapse_id = "navbar-collapse-" + private_random_int(1000, 10000)
            nav_container.append(
                tags.button(
                    tags.span(class_="navbar-toggler-icon"),
                    class_="navbar-toggler",
                    type="button",
                    data_bs_toggle="collapse",
                    data_bs_target="#" + collapse_id,
                    aria_controls=collapse_id,
                    aria_expanded="false",
                    aria_label="Toggle navigation",
                )
            )
            nav = div(nav, id=collapse_id, class_="collapse navbar-collapse")

        nav_container.append(nav)
        nav_final = tags.nav({"class": "navbar navbar-expand-md"}, nav_container)

        if self.position != "static-top":
            nav_final.add_class(self.position)

        nav_final.add_class(f"navbar-{'dark' if self.inverse else 'light'}")

        if self.bg:
            nav_final.attrs["style"] = "background-color: " + self.bg
        else:
            nav_final.add_class(f"bg-{'dark' if self.inverse else 'light'}")

        content = _make_tabs_fillable(
            content, self.fillable, gap=self.gap, padding=self.padding, navbar=True
        )

        # 2023-05-11; Do not wrap `row()` around `self.header` and `self.footer`
        contents: list[TagChild] = [
            child for child in [self.header, content, self.footer] if child is not None
        ]

        if self.sidebar is None:
            content_div = div(
                *contents,
                class_="container-fluid" if self.fluid else "container",
            )
            # If fillable is truthy, the .container also needs to be fillable
            if self.fillable:
                content_div = as_fill_carrier(content_div)
        else:
            content_div = div(
                # In the fluid case, the sidebar layout should be flush (i.e.,
                # the .container-fluid class adds padding that we don't want)
                {"class": "container"} if not self.fluid else None,
                layout_sidebar(
                    self.sidebar,
                    contents,
                    fillable=self.fillable is not False,
                    border_radius=False,
                    border=not self.fluid,
                ),
            )
            # Always have the sidebar layout fill its parent (in this case
            # fillable controls whether the _main_ content portion is fillable)
            content_div = as_fill_carrier(content_div)

        return TagList(nav_final, content_div)


# Given a .tab-content container, mark each relevant .tab-pane as a fill container/item.
def _make_tabs_fillable(
    content: Tag,
    fillable: bool | list[str] = True,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    navbar: bool = False,
) -> Tag:
    if not fillable:
        return content

    # Even if only one .tab-pane wants fillable behavior, the .tab-content
    # must to be a fillable container.
    content = as_fill_carrier(content)

    for child in content.children:
        # Only work on Tags
        if not isinstance(child, Tag):
            continue
        # Only work on .tab-pane children
        if not child.has_class("tab-pane"):
            continue
        # If `fillable` is a list, only fill the .tab-pane if its data-value is contained in `fillable`
        if isinstance(fillable, list):
            child_attr = child.attrs.get("data-value")
            if child_attr is None or child_attr not in fillable:
                continue
        styles = css(
            gap=as_css_unit(gap),
            padding=as_css_padding(padding),
            __bslib_navbar_margin="0;" if navbar else None,
        )
        child = tag_add_style(child, styles)
        child = as_fill_carrier(child)

    return content


def navset_bar(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    title: TagChild,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    fillable: bool | list[str] = True,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    position: Literal[
        "static-top", "fixed-top", "fixed-bottom", "sticky-top"
    ] = "static-top",
    header: TagChild = None,
    footer: TagChild = None,
    bg: Optional[str] = None,
    # TODO-bslib: default to 'auto', like we have in R (parse color via webcolors?)
    inverse: bool = False,
    collapsible: bool = True,
    fluid: bool = True,
) -> NavSetBar:
    """
    Render nav items as a navbar.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav`).
    title
        Title to display in the navbar.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the navigation elements into an expandable menu on mobile devices or narrow window widths.
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.

    See Also
    -------
    * ~shiny.ui.page_navbar
    * ~shiny.ui.nav
    * ~shiny.ui.nav_menu
    * ~shiny.ui.nav_control
    * ~shiny.ui.nav_spacer
    * ~shiny.ui.navset_tab
    * ~shiny.ui.navset_pill
    * ~shiny.experimental.ui.navset_tab_card
    * ~shiny.experimental.ui.navset_pill_card
    * ~shiny.ui.navset_hidden

    Example
    -------
    See :func:`~shiny.ui.nav`.
    """

    # If args contains any lists, flatten them into args.
    new_args: Sequence[NavSetArg | MetadataNode] = []
    for arg in args:
        if isinstance(arg, (list, tuple)):
            new_args.extend(arg)
        else:
            new_args.append(cast(NavSetArg, arg))

    return NavSetBar(
        *new_args,
        ul_class="nav navbar-nav",
        id=resolve_id(id) if id else None,
        selected=selected,
        sidebar=sidebar,
        fillable=fillable,
        gap=gap,
        padding=padding,
        title=title,
        position=position,
        header=header,
        footer=footer,
        bg=bg,
        inverse=inverse,
        collapsible=collapsible,
        fluid=fluid,
    )


# -----------------------------------------------------------------------------
# Utilities for rendering navs
# -----------------------------------------------------------------------------\
def _render_navset(
    *items: NavSetArg | MetadataNode,
    ul_class: str,
    id: Optional[str],
    selected: Optional[str],
    context: dict[str, Any],
) -> tuple[Tag, Tag]:
    tabsetid = private_random_int(1000, 10000)

    # Separate MetadataNodes from NavSetArgs.
    metadata_args = [x for x in items if isinstance(x, MetadataNode)]
    navset_args = [x for x in items if not isinstance(x, MetadataNode)]

    # If the user hasn't provided a selected value, use the first one
    if selected is None:
        for x in navset_args:
            selected = x.get_value()
            if selected is not None:
                break

    ul_tag = tags.ul(
        bootstrap_deps(),
        metadata_args,
        class_=ul_class,
        id=id,
        data_tabsetid=tabsetid,
    )
    div_tag = div(class_="tab-content", data_tabsetid=tabsetid)
    for i, x in enumerate(navset_args):
        nav, contents = x.resolve(
            selected, {**context, "tabsetid": tabsetid, "index": i}
        )
        ul_tag.append(nav)
        div_tag.append(contents)

    return ul_tag, div_tag


# # Card definition was gutted for bslib version.
# # * Bootstrap deps are not added

# def card(*args: TagChild, header: TagChild = None, footer: TagChild = None) -> Tag:
#     if header:
#         header = div(header, class_="card-header")
#     if footer:
#         footer = div(footer, class_="card-footer")

#     return div(
#         header,
#         div(*args, class_="card-body"),
#         footer,
#         bootstrap_deps(),
#         class_="card",
#     )



from __future__ import annotations

import htmltools

import shiny.experimental as x
from shiny import App, ui


def outer_inner() -> tuple[htmltools.Tag, htmltools.Tag, htmltools.Tag]:
    inner = ui.div(
        id="inner",
        style=htmltools.css(
            height="200px",
            border="3px blue solid",
        ),
    )
    middle = ui.div(
        inner,
        id="middle",
        style=htmltools.css(
            height="300px",
            border="3px green solid",
        ),
    )
    outer = ui.div(
        middle,
        id="outer",
        style=htmltools.css(
            height="400px",
            border="3px red solid",
        ),
    )
    return outer, middle, inner


# outer0, inner0 = outer_inner()

outer0, middle0, inner0 = outer_inner()
outer1, middle1, inner1 = outer_inner()
outer2, middle2, inner2 = outer_inner()
outer3, middle3, inner3 = outer_inner()
x.ui.as_fillable_container(outer0)
x.ui.as_fillable_container(outer1)
x.ui.as_fillable_container(outer2)
x.ui.as_fillable_container(outer3)

x.ui.as_fill_item(inner0)
x.ui.as_fill_item(inner1)
x.ui.as_fill_item(inner2)
x.ui.as_fill_item(inner3)

x.ui.as_fill_item(middle1)
x.ui.as_fillable_container(middle2)
x.ui.as_fill_carrier(middle3)

app_ui = ui.page_fluid(
    ui.markdown(
        """\
        # `as_fill_carrier()`

        For an element to pass through the ability to fill layout,
        * the element must have `as_fill_item(el)` be called to allow it to expand
        * the element must have `as_fillable_container(el)` called on it to allow its contents to expand

        This can be shortened by calling `as_fill_carrier(el)`.

        For all examples below,
        * `as_fillable_container(red)` has been called, allowing its contents to naturally expand.
        * `as_fill_item(blue)` has been called, allowing it to expand into its parent container
        """
    ),
    ui.row(
        ui.column(3, ui.h5("Default")),
        ui.column(3, ui.h5(ui.markdown("`as_fill_item(green)`"))),
        ui.column(3, ui.h5(ui.markdown("`as_fillable_container(green)`"))),
        ui.column(3, ui.h5(ui.markdown("`as_fill_carrier(green)`"))),
    ),
    ui.row(
        ui.column(3, ui.div(outer0)),
        ui.column(3, ui.div(outer1)),
        ui.column(3, ui.div(outer2)),
        ui.column(3, ui.div(outer3)),
    ),
)

app = App(app_ui, server=None)



from __future__ import annotations

from typing import TypeVar

from htmltools import Tag

TagT = TypeVar("TagT", bound="Tag")

# Tag.add_class(x: str) -> Self[Tag]:
#     cls = self.attrs.get("class")
#     if cls:
#         x = cls + " " + x
#     self.attrs["class"] = x
#     return self


# Do not export
# "Prepend" version of `tag.add_class(class_)`
def tag_prepend_class(tag: TagT, *class_: str | None) -> TagT:
    classes = (
        *class_,
        tag.attrs.get("class"),
    )
    classes = [c for c in classes if c is not None]
    if len(classes) == 0:
        return tag
    tag.attrs["class"] = " ".join(classes)
    return tag


def tag_remove_class(tag: TagT, *class_: str | None) -> TagT:
    """
    Remove a class value from the HTML class attribute.

    Parameters
    ----------
    *class_
        The class name to remove.

    Returns
    -------
    :
        The modified tag.
    """
    cls = tag.attrs.get("class")
    # If no class values to remove from, quit
    if not cls:
        return tag

    # Remove any `None` values
    set_to_remove = {c for c in class_ if c is not None}

    # If no classes to remove, quit
    if len(set_to_remove) == 0:
        return tag

    # Get new set of classes
    # Order matters, otherwise we could use `set()` subtraction: `set(cls.split(" ")) - set(class_)`
    new_cls: list[str] = []
    for cls_val in cls.split(" "):
        if cls_val not in set_to_remove:
            new_cls.append(cls_val)

    # If no classes left, remove the attribute
    if len(new_cls) == 0:
        # If here, `attrs.class` must exist
        tag.attrs.pop("class")
        return tag

    # Otherwise, set the new class
    tag.attrs["class"] = " ".join(new_cls)
    return tag


def tag_add_style(
    tag: TagT,
    *style: str | None,
) -> TagT:
    """
    Add a style value(s) to the HTML style attribute.

    Parameters
    ----------
    *style
        CSS properties and values already properly formatted. Each should already contain trailing semicolons.

    See Also
    --------
    ~htmltools.css

    Returns
    -------
    :
        The modified tag.
    """
    styles = (
        tag.attrs.get("style"),
        *style,
    )
    non_none_style_tuple = (s for s in styles if s is not None)
    style_str = "".join(non_none_style_tuple)

    if style_str:
        tag.attrs["style"] = style_str
    return tag



from __future__ import annotations

import base64
import io
import mimetypes
from pathlib import Path, PurePath
from typing import Literal, Optional, Protocol

from htmltools import (
    HTML,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    Tagifiable,
    TagList,
    css,
    div,
    tags,
)

from ...types import MISSING, MISSING_TYPE
from ._css_unit import CssUnit, as_css_padding, as_css_unit
from ._fill import as_fill_carrier, as_fill_item, as_fillable_container
from ._htmldeps import card_dependency
from ._tooltip import tooltip
from ._utils import consolidate_attrs

__all__ = (
    "CardItem",
    "card",
    "card_body",
    "card_title",
    "card_header",
    "card_footer",
    "card_image",
)


# TODO-maindocs; @add_example()
def card(
    *args: TagChild | TagAttrs | CardItem,
    full_screen: bool = False,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    wrapper: WrapperCallable | None | MISSING_TYPE = MISSING,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    A Bootstrap card component

    A general purpose container for grouping related UI elements together with a border
    and optional padding. To learn more about `card()`s, see [this
    article](https://rstudio.github.io/bslib/articles/cards.html).

    Parameters
    ----------
    *args
        Unnamed arguments can be any valid child of an :class:`~htmltools.Tag` (which
        includes card items such as :func:`~shiny.experimental.ui.card_body`. Named
        arguments become HTML attributes on the returned Tag.
    full_screen
        If `True`, an icon will appear when hovering over the card body. Clicking the
        icon expands the card to fit viewport size.
    height,max_height,min_height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`).
    fill
        Whether or not to allow the card to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.experimental.ui.page_fillable`).
    class_
        Additional CSS classes for the returned Tag.
    wrapper
        A function (which returns a UI element) to call on unnamed arguments in `*args`
        which are not already card item(s) (like
        :func:`~shiny.experimental.ui.card_header`,
        :func:`~shiny.experimental.ui.card_body`, etc.). Note that non-card items are
        grouped together into one `wrapper` call (e.g. given `card("a", "b",
        card_body("c"), "d")`, `wrapper` would be called twice, once with `"a"` and
        `"b"` and once with `"d"`).

    Returns
    -------
    :
        An :func:`~htmltools.div` tag.

    See Also
    --------
    * :func:`~shiny.experimental.ui.layout_column_wrap` for laying out multiple cards
      (or multiple columns inside a card).
    * :func:`~shiny.experimental.ui.card_header` for creating a header within the card.
    * :func:`~shiny.experimental.ui.card_title` for creating a title within the card body.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside the card.
    * :func:`~shiny.experimental.ui.card_footer` for creating a footer within the card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to the card.
    """
    if isinstance(wrapper, MISSING_TYPE):
        wrapper = card_body

    attrs, children = consolidate_attrs(*args, class_=class_, **kwargs)
    children = _wrap_children_in_card(*children, wrapper=wrapper)

    tag = div(
        {
            "class": "card bslib-card bslib-mb-spacer",
            "style": css(
                height=as_css_unit(height),
                max_height=as_css_unit(max_height),
                min_height=as_css_unit(min_height),
            ),
            "data-bslib-card-init": True,
            "data-full-screen": "false" if full_screen else None,
        },
        *children,
        attrs,
        _full_screen_toggle() if full_screen else None,
        card_dependency(),
        _card_js_init(),
    )
    tag = as_fillable_container(tag)
    if fill:
        tag = as_fill_item(tag)

    return tag


def _card_js_init() -> Tag:
    return tags.script(
        {"data-bslib-card-init": True},
        "window.bslib.Card.initializeAllCards();",
    )


def _full_screen_toggle() -> Tag:
    return tooltip(
        tags.span(
            {"class": "bslib-full-screen-enter badge rounded-pill bg-dark"},
            _full_screen_toggle_icon(),
        ),
        "Expand",
    )


# via bsicons::bs_icon("arrows-fullscreen")
def _full_screen_toggle_icon() -> HTML:
    return HTML(
        '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16" class="bi bi-arrows-fullscreen " style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img" ><path fill-rule="evenodd" d="M5.828 10.172a.5.5 0 0 0-.707 0l-4.096 4.096V11.5a.5.5 0 0 0-1 0v3.975a.5.5 0 0 0 .5.5H4.5a.5.5 0 0 0 0-1H1.732l4.096-4.096a.5.5 0 0 0 0-.707zm4.344 0a.5.5 0 0 1 .707 0l4.096 4.096V11.5a.5.5 0 1 1 1 0v3.975a.5.5 0 0 1-.5.5H11.5a.5.5 0 0 1 0-1h2.768l-4.096-4.096a.5.5 0 0 1 0-.707zm0-4.344a.5.5 0 0 0 .707 0l4.096-4.096V4.5a.5.5 0 1 0 1 0V.525a.5.5 0 0 0-.5-.5H11.5a.5.5 0 0 0 0 1h2.768l-4.096 4.096a.5.5 0 0 0 0 .707zm-4.344 0a.5.5 0 0 1-.707 0L1.025 1.732V4.5a.5.5 0 0 1-1 0V.525a.5.5 0 0 1 .5-.5H4.5a.5.5 0 0 1 0 1H1.732l4.096 4.096a.5.5 0 0 1 0 .707z"></path></svg>'
    )


############################################################################


class CardItem:
    """
    A wrapper around a :class:`~htmltools.Tag` object that represents a card item (e.g.,
    :func:`~shiny.experimental.ui.card_body`,
    :func:`~shiny.experimental.ui.card_header`, etc.).

    This class is used to allow for consecutive non-card items to be bundled into a
    single `card_body` within :func:`~shiny.experimental.ui.card`.

    Parameters
    ----------
    item
        A :class:`~htmltools.Tag` object that represents a card item (e.g.,
        :func:`~shiny.experimental.ui.card_body`, :func:`~shiny.experimental.ui.card_header`, etc.).

    See Also
    --------
    * :func:`~shiny.experimental.ui.card` for creating a card component.
    * :func:`~shiny.experimental.ui.card_header` for creating a header within the card.
    * :func:`~shiny.experimental.ui.card_title` for creating a title within the card body.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside the card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to the card.
    * :func:`~shiny.experimental.ui.card_footer` for creating a footer within the card.
    """

    def __init__(
        self,
        item: TagChild,
    ):
        self._item = item

    def resolve(self) -> TagChild:
        """
        Resolves the `CardItem` class by returning the `item` provided at initialization.

        Returns
        -------
        :
            The `item` provided at initialization.
        """
        return self._item

    def tagify(self) -> TagList:
        """
        Tagify the `item`

        Returns
        -------
        :
            A tagified :class:`~htmltools.TagList` object.
        """
        return TagList(self.resolve()).tagify()


# TODO-maindocs; @add_example()
def card_body(
    *args: TagChild | TagAttrs,
    fillable: bool = True,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    max_height_full_screen: Optional[CssUnit] | MISSING_TYPE = MISSING,
    height: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> CardItem:
    # For a general overview of the :func:`~shiny.experimental.ui.card` API, see [this article](https://rstudio.github.io/bslib/articles/cards.html).
    """
    Card body container

    A general container for the "main content" of a :func:`~shiny.experimental.ui.card`. This component is designed
    to be provided as direct children to :func:`~shiny.experimental.ui.card`.

    Parameters
    ----------
    *args
        Contents to the card's body. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    fillable
        Whether or not the card item should be a fillable (i.e. flexbox) container.
    min_height,max_height,max_height_full_screen
        Any valid CSS length unit. If `max_height_full_screen` is missing, it is set to
        `max_height`.
    height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`).
    padding
        Padding to use for the body. This can be a numeric vector
        (which will be interpreted as pixels) or a character vector with valid CSS
        lengths. The length can be between one and four. If one, then that value
        will be used for all four sides. If two, then the first value will be used
        for the top and bottom, while the second value will be used for left and
        right. If three, then the first will be used for top, the second will be
        left and right, and the third will be bottom. If four, then the values will
        be interpreted as top, right, bottom, and left respectively.
    gap
        A CSS length unit defining the `gap` (i.e., spacing) between elements provided
        to `*args`. This argument is only applicable when `fillable = TRUE`.
    fill
        Whether to allow this element to grow/shrink to fit its `card` container.
    class_
        Additional CSS classes for the returned Tag.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.experimental.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.experimental.ui.layout_column_wrap` for laying out multiple cards
        (or multiple columns inside a card).
    * :func:`~shiny.experimental.ui.card` for creating a card component.
    * :func:`~shiny.experimental.ui.card_header` for creating a header within the card.
    * :func:`~shiny.experimental.ui.card_title` for creating a title within the card body.
    * :func:`~shiny.experimental.ui.card_footer` for creating a footer within the card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to the card.
    """
    if isinstance(max_height_full_screen, MISSING_TYPE):
        max_height_full_screen = max_height

    div_style_args = {
        "min-height": as_css_unit(min_height),
        "--bslib-card-body-max-height": as_css_unit(max_height),
        "--bslib-card-body-max-height-full-screen": as_css_unit(max_height_full_screen),
        "margin-top": "auto",
        "margin-bottom": "auto",
        # .card-body already adds `flex: 1 1 auto` so make sure to override it
        "flex": "1 1 auto" if fill else "0 0 auto",
        "padding": as_css_padding(padding),
        "gap": as_css_unit(gap),
        "height": as_css_unit(height),
    }
    tag = tags.div(
        {
            "class": "card-body bslib-gap-spacing",
            "style": css(**div_style_args),
        },
        *args,
        class_=class_,
        **kwargs,
    )

    if fillable:
        tag = as_fillable_container(tag)
    if fill:
        tag = as_fill_item(tag)

    return CardItem(tag)


# https://mypy.readthedocs.io/en/stable/protocols.html#callback-protocols
class WrapperCallable(Protocol):
    """
    A callable that wraps children into a :func:`~shiny.experimental.ui.CardItem`.
    """

    def __call__(self, *args: TagChild) -> CardItem:
        """
        Wraps children into a :func:`~shiny.experimental.ui.CardItem`.

        Parameters
        ----------
        *args
            `TagChild` children to wrap.

        Returns
        -------
        :
            A :class:`~shiny.experimental.ui.CardItem` object.
        """
        ...


def _as_card_items(
    *children: TagChild | CardItem | None,  # `TagAttrs` are not allowed here
    wrapper: WrapperCallable | None,
) -> list[CardItem]:
    # We don't want `None`s creating empty card bodies
    children_vals = [child for child in children if child is not None]

    attrs, children_vals = consolidate_attrs(*children_vals)
    if len(attrs) > 0:
        raise ValueError("`TagAttrs` are not allowed in `_as_card_items(*children=)`.")

    if not callable(wrapper):
        ret: list[CardItem] = []
        for child in children_vals:
            if isinstance(child, CardItem):
                ret.append(child)
            else:
                ret.append(CardItem(child))
        return ret

    # Any children that are `is.card_item` should be included verbatim. Any
    # children that are not, should be wrapped in card_body(). Consecutive children
    # that are not card_item, should be wrapped in a single card_body().
    state = "asis"  # "wrap" | "asis"
    new_children: list[CardItem] = []
    children_to_wrap: list[TagChild] = []

    def wrap_children():
        nonlocal children_to_wrap
        wrapped_children = wrapper(*children_to_wrap)
        new_children.append(wrapped_children)
        children_to_wrap = []

    for child in children_vals:
        if isinstance(child, CardItem):
            if state == "wrap":
                wrap_children()
            state = "asis"
            new_children.append(child)
        else:
            # Not a card, collect it for wrapping
            state = "wrap"
            children_to_wrap.append(child)
    if state == "wrap":
        wrap_children()

    return new_children


def _wrap_children_in_card(
    *children: TagChild | CardItem | None,  # `TagAttrs` are not allowed here
    wrapper: WrapperCallable | None,
) -> list[TagChild]:
    card_items = _as_card_items(*children, wrapper=wrapper)
    tag_children = [card_item.resolve() for card_item in card_items]
    return tag_children


# https://mypy.readthedocs.io/en/stable/protocols.html#callback-protocols
class TagCallable(Protocol):  # Should this be exported from htmltools?
    """
    Callable definition for a defined :class:`~htmltools.Tag` method.

    This is used to define the `container` argument in :func:`~shiny.experimental.ui.card_title`,
    :func:`~shiny.experimental.ui.card_header`, and :func:`~shiny.experimental.ui.card_footer`.

    See Also
    --------
    * :class:`~htmltools.Tag`
    """

    def __call__(
        self,
        *args: TagChild | TagAttrs,
        _add_ws: bool = True,
        **kwargs: TagAttrValue,
    ) -> Tagifiable:
        """
        A tag method.

        Parameters
        ----------
        *args
            Contents to the tag method. Or tag attributes that are supplied to the
            resolved :class:`~htmltools.Tag` object.
        _add_ws
            Whether or not to add whitespace to the returned :class:`~htmltools.Tag`
            object.
        **kwargs
            Additional HTML attributes for the returned Tag.

        Returns
        -------
        :
            A :class:`~htmltools.Tag` object.
        """
        ...


# TODO-maindocs; @add_example()
def card_title(
    *args: TagChild | TagAttrs,
    container: TagCallable = tags.h5,
    **kwargs: TagAttrValue,
) -> Tagifiable:
    """
    Card title container

    A general container for the "title" of a :func:`~shiny.experimental.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.experimental.ui.card`.

    Parameters
    ----------
    *args
        Contents to the card's title. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    container
        Method for the returned Tag object. Defaults to :func:`shiny.ui.tags.h5`.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A Tag object.

    See Also
    --------
    * :func:`~shiny.experimental.ui.card` for creating a card component.
    * :func:`~shiny.experimental.ui.card_header` for creating a header within the card.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside the card.
    * :func:`~shiny.experimental.ui.card_footer` for creating a footer within the card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to the card.
    """
    return container(*args, **kwargs)


# TODO-maindocs; @add_example()
def card_header(
    *args: TagChild | TagAttrs,
    container: TagCallable = tags.div,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Card header container

    A general container for the "header" of a :func:`~shiny.experimental.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.experimental.ui.card`.

    The header has a different background color and border than the rest of the card.

    Parameters
    ----------
    *args
        Contents to the header container. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    container
        Method for the returned Tag object. Defaults to :func:`~shiny.ui.tags.div`.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.experimental.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.experimental.ui.card` for creating a card component.
    * :func:`~shiny.experimental.ui.card_title` for creating a title within the card body.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside the card.
    * :func:`~shiny.experimental.ui.card_footer` for creating a footer within the card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to the card.
    """
    return CardItem(
        container({"class": "card-header"}, *args, **kwargs),
    )


# TODO-maindocs; @add_example()
def card_footer(
    *args: TagChild | TagAttrs,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Card footer container

    A general container for the "footer" of a :func:`~shiny.experimental.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.experimental.ui.card`.

    The footer has a different background color and border than the rest of the card.

    Parameters
    ----------
    *args
        Contents to the footer container. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.experimental.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.experimental.ui.card` for creating a card component.
    * :func:`~shiny.experimental.ui.card_title` for creating a title within the card body.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside the card.
    * :func:`~shiny.experimental.ui.card_footer` for creating a footer within the card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to the card.
    """
    return CardItem(
        tags.div({"class": "card-footer"}, *args, **kwargs),
    )


class ImgContainer(Protocol):
    """
    A callable that wraps the return value of `card_image()`. To isolate your object in a card, return a :func:`~shiny.experimental.ui.CardItem`.
    """

    def __call__(self, *args: Tag) -> Tagifiable:
        """
        Wraps the return value of `card_image()`.

        Parameters
        ----------
        *args
            The return value of `card_image()`.

        Returns
        -------
        :
            A tagifiable object, such as a :class:`~htmltools.Tag` or
            :class:`~shiny.experimental.ui.CardItem` object.
        """
        ...


# TODO-maindocs; @add_example()
def card_image(
    file: str | Path | PurePath | io.BytesIO | None,
    *args: TagAttrs,
    href: Optional[str] = None,
    border_radius: Literal["top", "bottom", "all", "none"] = "top",
    mime_type: Optional[str] = None,
    class_: Optional[str] = None,
    height: Optional[CssUnit] = None,
    fill: bool = True,
    width: Optional[CssUnit] = None,
    # Required so that multiple `card_images()` are not put in the same `card()`
    container: ImgContainer = card_body,
    **kwargs: TagAttrValue,
) -> Tagifiable:
    """
    Card image container

    A general container for an image within a :func:`~shiny.experimental.ui.card`. This component is designed to be
    provided as a direct child to :func:`~shiny.experimental.ui.card`.

    Parameters
    ----------
    file
        A file path pointing an image. The image will be base64 encoded and provided to
        the `src` attribute of the `<img>`. Alternatively, you may set this value to
        `None` and provide the `src` yourself via `*args:TagAttrs` or
        `**kwargs:TagAttrValue` (e.g. `{"src": "HOSTED_PATH_TO_IMAGE"}` or
        `src="HOSTED_PATH_TO_IMAGE"`).
    *args
        Dictionary of tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    href
        An optional URL to link to.
    border_radius
        Where to apply `border-radius` on the image.
    mime_type
        The mime type of the `file`.
    class_
        Additional CSS classes for the resolved Tag.
    height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`).
    fill
        Whether to allow this element to grow/shrink to fit its `card` container.
    width
        Any valid CSS unit (e.g., `width="100%"`).
    container
        Method to wrap the returned Tag object. Defaults to :func:`~shiny.experimental.ui.card_body`.
        If :func:`~shiny.experimental.ui.card_body` is used, each image will be in separate cards. If
        the `container` method does not return a :class:`~shiny.experimental.ui.CardItem`, it
        allows for consecutive non-`CardItem` objects to be bundled into a single
        :func:`~.shiny.experimental.card_body` within :func:`~shiny.experimental.ui.card`.
    **kwargs
        Additional HTML attributes for the resolved Tag.
    """
    src = None
    if file is not None:
        if isinstance(file, io.BytesIO):
            b64_str = base64.b64encode(file.read()).decode("utf-8")
            if mime_type is None:
                raise ValueError(
                    "`mime_type` must be provided when passing an in-memory buffer"
                )
            src = f"data:{mime_type};base64,{b64_str}"

        elif isinstance(file, (str, Path, PurePath)):
            with open(file, "rb") as img_file:
                b64_str = base64.b64encode(img_file.read()).decode("utf-8")
                if mime_type is None:
                    mime_type = mimetypes.guess_type(file)[0]
                src = f"data:{mime_type};base64,{b64_str}"

    card_class_map = {
        "all": "card-img",
        "top": "card-img-top",
        "bottom": "card-img-bottom",
    }

    image = tags.img(
        {
            "src": src,
            "class": "img-fluid",
            "style": css(
                height=as_css_unit(height),
                width=as_css_unit(width),
            ),
        },
        {"class": card_class_map.get(border_radius, None)},
        *args,
        class_=class_,
        **kwargs,
    )

    if fill:
        image = as_fill_item(image)

    if href is not None:
        image = as_fill_carrier(
            tags.a(
                image,
                href=href,
            )
        )

    if container:
        return container(image)
    else:
        return CardItem(image)



from __future__ import annotations

import shiny.experimental as x
from shiny import App, ui
from shiny.types import NavSetArg

my_sidebar = x.ui.sidebar("My sidebar!!", open="open", title="Barret")


def nav_with_content(letter: str, prefix: str) -> ui._navs.Nav:
    return ui.nav(letter, ui.markdown(f"`{prefix}`: tab {letter} content"))


def nav_items(prefix: str) -> list[NavSetArg]:
    a = nav_with_content("a", prefix)
    b = nav_with_content("b", prefix)
    github = ui.nav_control(
        ui.tags.a(
            # ui.icon("github"),
            "Shiny",
            href="https://github.com/rstudio/shiny",
            target="_blank",
        ),
    )
    space = ui.nav_spacer()
    other = ui.nav_menu(
        "Other links",
        nav_with_content("c", prefix),
        ui.nav_control(
            ui.tags.a(
                # icon("r-project"),
                "RStudio",
                href="https://rstudio.com",
                target="_blank",
            ),
        ),
        align="right",
    )
    return [a, b, github, space, other]


app = App(
    ui=x.ui.page_navbar(
        # theme = bs_theme(),
        *nav_items("page_navbar()"),
        sidebar=my_sidebar,
        title="page_navbar()",
        bg="#0062cc",
        header=ui.markdown(
            "Testing app for `bslib::nav_spacer()` and `bslib::nav_item()` [#319](https://github.com/rstudio/bslib/pull/319)."
        ),
        footer=ui.div(
            {"style": "width:80%; margin: 0 auto"},
            ui.h4("navset_tab_card()"),
            x.ui.navset_tab_card(
                *nav_items("navset_tab_card()"),
                sidebar=my_sidebar,
            ),
            ui.h4("navset_pill_card()"),
            x.ui.navset_pill_card(
                *nav_items("navset_pill_card()"),
                sidebar=my_sidebar,
            ),
            # Do not include `navset_bar()` in example. Ok for testing only
            ui.h4("navset_bar()"),
            x.ui.navset_bar(
                *nav_items("navset_bar()"),
                title="Test!",
                sidebar=my_sidebar,
            ),
        ),
    ),
    server=None,
)



from __future__ import annotations

from typing import TypeVar, overload

from ._docstring import add_example
from .types import SilentCancelOutputException, SilentException

T = TypeVar("T")


@overload
def req(*, cancel_output: bool = False) -> None:
    ...


@overload
def req(*args: T, cancel_output: bool = False) -> T:
    ...


@add_example()
def req(*args: T, cancel_output: bool = False) -> T | None:
    """
    Throw a silent exception for falsy values.

    This is a convenient shorthand for throwing :func:`~shiny.types.SilentException` /
    :func:`~shiny.types.SilentCancelOutputException` if any of the arguments are falsy.

    Parameters
    ----------
    args
        Any number of arguments to check.
    cancel_output
        If ``True``, throw :func:`~shiny.types.SilentCancelOutputException` instead of
        :func:`~shiny.types.SilentException`.

    Returns
    -------
    :
        The first argument. If no arguments are provided, returns ``None``.
    """
    if len(args) == 0:
        return None

    for arg in args:
        if not arg:
            if cancel_output:
                raise SilentCancelOutputException()
            else:
                raise SilentException()

    return args[0]



from __future__ import annotations

import shiny.experimental as x
from shiny import App, ui

y = x.ui.card("A simple card")

app_ui = ui.page_fluid(
    # Always has 2 columns (on non-mobile)
    x.ui.layout_column_wrap(1 / 2, y, y, y),
    ui.hr(),
    # Has three columns when viewport is wider than 750px
    x.ui.layout_column_wrap("250px", y, y, y),
)


app = App(app_ui, server=None)



import shiny.experimental as x
from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    x.ui.input_text_area("caption", "Caption:", "Data summary", autoresize=True),
    ui.output_text_verbatim("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @output
    @render.text
    def value():
        return input.caption()


app = App(app_ui, server)



# Needed for types imported only during TYPE_CHECKING with Python 3.7 - 3.9
# See https://www.python.org/dev/peps/pep-0655/#usage-in-python-3-11
from __future__ import annotations

import base64
import io
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, cast

from ..types import ImgData, PlotnineFigure
from ._coordmap import get_coordmap, get_coordmap_plotnine

TryPlotResult = Tuple[bool, "ImgData| None"]


if TYPE_CHECKING:
    from matplotlib.figure import Figure


# Try to render a matplotlib object (or the global figure, if it's been used). If `fig`
# is not a matplotlib object, return (False, None). If there's an error in rendering,
# return None. If successful in rendering, return an ImgData object.
def try_render_matplotlib(
    x: object,
    width: float,
    height: float,
    pixelratio: float,
    ppi: float,
    allow_global: bool,
    alt: Optional[str],
    **kwargs: object,
) -> TryPlotResult:
    fig = get_matplotlib_figure(x, allow_global)

    if fig is None:
        return (False, None)

    try:
        import matplotlib.pyplot as plt

        fig.set_size_inches(width / ppi, height / ppi)
        fig.set_dpi(ppi * pixelratio)

        plt.tight_layout()  # pyright: ignore[reportUnknownMemberType]
        coordmap = get_coordmap(fig)

        with io.BytesIO() as buf:
            fig.savefig(  # pyright: ignore[reportUnknownMemberType]
                buf,
                format="png",
                dpi=ppi * pixelratio,
                **kwargs,
            )
            buf.seek(0)
            data = base64.b64encode(buf.read())
            data_str = data.decode("utf-8")

        res: ImgData = {
            "src": "data:image/png;base64," + data_str,
            "width": "100%",
            "height": "100%",
        }

        if alt is not None:
            res["alt"] = alt

        if coordmap is not None:
            res["coordmap"] = coordmap

        return (True, res)

    finally:
        import matplotlib.pyplot

        matplotlib.pyplot.close(fig)  # pyright: ignore[reportUnknownMemberType]


def get_matplotlib_figure(
    x: object, allow_global: bool
) -> Figure | None:  # pyright: ignore
    import matplotlib.pyplot as plt
    from matplotlib.animation import Animation
    from matplotlib.artist import Artist
    from matplotlib.figure import Figure

    # Detect usage of pyplot global figure
    # TODO: Might be good to detect non-empty plt.get_fignums() before we call the user
    #   function, which would mean we will false-positive here. Maybe we warn in that
    #   case, maybe we ignore gcf(), maybe both.
    if x is None and len(plt.get_fignums()) > 0:
        if allow_global:
            return plt.gcf()
        else:
            # Must close the global figure so we don't stay in this state forever
            plt.close(plt.gcf())  # pyright: ignore[reportUnknownMemberType]
            raise RuntimeError(
                "matplotlib.pyplot cannot be used from an async render function; "
                "please use matplotlib's object-oriented interface instead"
            )

    if isinstance(x, Figure):
        return x

    if isinstance(x, Animation):
        raise RuntimeError(
            "Matplotlib's Animation class isn't supported by @render.plot. "
            + "Consider explictly saving the animation to a file and "
            + "then using @render.image instead to render it."
        )

    # Libraries like pandas, xarray, etc have plot() methods that can return a wide
    # array pf mpl classes, like Lines2D, Subplots, Axes, etc. The Artist ABC class
    # should cover most, if not all, of these (it doesn't cover Animation, though).
    # https://matplotlib.org/stable/api/artist_api.html
    if isinstance(x, Artist):
        return x.get_figure()

    # Some other custom figure-like classes such as seaborn.axisgrid.FacetGrid attach
    # their figure as an attribute
    fig = getattr(x, "figure", None)
    if isinstance(fig, Figure):
        return fig

    # Sometimes generic plot() methods will return an iterable of Artists,
    # If they all refer to the same figure, then it seems reasonable to use it
    # https://docs.xarray.dev/en/latest/user-guide/plotting.html#dimension-along-y-axis
    if isinstance(x, (list, tuple)):
        figs = [get_matplotlib_figure(y, allow_global) for y in cast(List[Any], x)]
        if len(set(figs)) == 1:
            return figs[0]

    return None


def try_render_pil(
    x: object,
    width: float,
    height: float,
    pixelratio: float,
    ppi: float,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    import PIL.Image

    if not isinstance(x, PIL.Image.Image):
        return (False, None)

    with io.BytesIO() as buf:
        x.save(buf, format="PNG", **kwargs)
        buf.seek(0)
        data = base64.b64encode(buf.read())
        data_str = data.decode("utf-8")

    res: ImgData = {
        "src": "data:image/png;base64," + data_str,
        "width": "100%",
        "height": "100%",
        "style": "object-fit:contain",
    }

    if alt is not None:
        res["alt"] = alt

    return (True, res)


def try_render_plotnine(
    x: object,
    width: float,
    height: float,
    pixelratio: float,
    ppi: float,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    from plotnine.ggplot import ggplot

    if not isinstance(x, ggplot):
        return (False, None)

    x = cast(PlotnineFigure, x)

    with io.BytesIO() as buf:
        # save_helper was added in plotnine 0.10.1-dev. If this method exists, we can
        # use it to get the matplotlib Figure object, which we can then use to get the
        # coordmap. Once this version of plotnine is released and in common use, we can
        # add a version dependency and remove the conditional code.
        if hasattr(x, "save_helper"):
            res = x.save_helper(  # pyright: ignore[reportUnknownMemberType, reportGeneralTypeIssues, reportUnknownVariableType]
                filename=buf,
                format="png",
                units="in",
                dpi=ppi * pixelratio,
                width=width / ppi,
                height=height / ppi,
                verbose=False,
                **kwargs,
            )
            coordmap = get_coordmap_plotnine(
                x,
                res.figure,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType, reportGeneralTypeIssues]
            )
            res.figure.savefig(  # pyright: ignore[reportUnknownMemberType, reportGeneralTypeIssues]
                **res.kwargs  # pyright: ignore[reportUnknownMemberType, reportGeneralTypeIssues]
            )
        else:
            x.save(
                filename=buf,
                format="png",
                units="in",
                dpi=ppi * pixelratio,
                width=width / ppi,
                height=height / ppi,
                verbose=False,
                **kwargs,
            )
            coordmap = None
        buf.seek(0)
        data = base64.b64encode(buf.read())
        data_str = data.decode("utf-8")

    res: ImgData = {
        "src": "data:image/png;base64," + data_str,
        "width": "100%",
        "height": "100%",
    }

    if alt is not None:
        res["alt"] = alt

    if coordmap is not None:
        res["coordmap"] = coordmap

    return (True, res)



from __future__ import annotations

import random
from typing import Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, TagList, css, div
from htmltools import svg as svgtags
from htmltools import tags

from ... import Session
from ..._deprecated import warn_deprecated
from ...session import require_active_session

# from ._color import get_color_contrast
from ._card import CardItem
from ._css_unit import CssUnit, as_css_padding, as_css_unit
from ._fill import as_fill_item, as_fillable_container
from ._htmldeps import sidebar_dependency
from ._utils import consolidate_attrs, trinary


class Sidebar:
    """
    Sidebar object

    Class returned from :func:`~shiny.experimental.ui.sidebar`. Please do not use this
    class directly. Instead, supply the :func:`~shiny.experimental.ui.sidebar` object to
    :func:`~shiny.experimental.ui.layout_sidebar`.

    Attributes
    ----------
    tag
        The :class:`~htmltools.Tag` object that represents the sidebar.
    collapse_tag
        The :class:`~htmltools.Tag` object that represents the collapse button.
    position
        Where the sidebar should appear relative to the main content.
    open
        The initial state of the sidebar.
    width
        A valid CSS unit used for the width of the sidebar.
    max_height_mobile
        The maximum height of the horizontal sidebar when viewed on mobile devices.
        The default is `250px` unless the sidebar is included in a
        :func:`~shiny.experimental.ui.layout_sidebar` with a specified height, in
        which case the default is to take up no more than 50% of the layout container.
    color_fg
        A foreground color.
    color_bg
        A background color.

    Parameters
    ----------
    tag
        The :class:`~htmltools.Tag` object that represents the sidebar.
    collapse_tag
        The :class:`~htmltools.Tag` object that represents the collapse button.
    position
        Where the sidebar should appear relative to the main content.
    open
        The initial state of the sidebar.
    width
        A valid CSS unit used for the width of the sidebar.
    max_height_mobile
        The maximum height of the horizontal sidebar when viewed on mobile devices.
        The default is `250px` unless the sidebar is included in a
        :func:`~shiny.experimental.ui.layout_sidebar` with a specified height, in
        which case the default is to take up no more than 50% of the layout container.
    color_fg
        A foreground color.
    color_bg
        A background color.


    """

    def __init__(
        self,
        tag: Tag,
        collapse_tag: Optional[Tag],
        position: Literal["left", "right"],
        open: Literal["desktop", "open", "closed", "always"],
        width: CssUnit,
        max_height_mobile: Optional[str | float],
        color_fg: Optional[str],
        color_bg: Optional[str],
    ):
        self.tag = tag
        self.collapse_tag = collapse_tag
        self.position = position
        self.open = open
        self.width = width
        self.max_height_mobile = max_height_mobile
        self.color_fg = color_fg
        self.color_bg = color_bg

    # The `Sidebar` class should use it's fields, not this method
    def tagify(self) -> Tag:
        """
        Not implemented
        """
        # Similar to `NavMenu.tagify()`
        raise NotImplementedError(
            "`Sidebar` objects must be handled by `layout_sidebar(sidebar)`."
        )


# TODO-maindocs; @add_example()
def sidebar(
    *args: TagChild | TagAttrs,
    width: CssUnit = 250,
    position: Literal["left", "right"] = "left",
    open: Literal["desktop", "open", "closed", "always"] = "desktop",
    id: Optional[str] = None,
    title: TagChild | str = None,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    class_: Optional[str] = None,  # TODO-future; Consider using `**kwargs` instead
    max_height_mobile: Optional[str | float] = None,
) -> Sidebar:
    # See [this article](https://rstudio.github.io/bslib/articles/sidebars.html)
    #   to learn more.
    # TODO-future; If color contrast is implemented. Docs for `bg` and `fg`:
    #     If only one of either is provided, an
    #     accessible contrasting color is provided for the opposite color, e.g. setting
    #     `bg` chooses an appropriate `fg` color.

    """
    Sidebar element

    Create a collapsing sidebar layout by providing a `sidebar()` object to the
    `sidebar=` argument of:

    * :func:`~shiny.experimental.ui.layout_sidebar`
      * Creates a sidebar layout component which can be dropped inside any
        :func:`~shiny.ui.page` or :func:`~shiny.experimental.ui.card` context.
    * :func:`~shiny.experimental.ui.navset_bar`, :func:`~shiny.experimental.ui.navset_tab_card`, and :func:`~shiny.experimental.ui.navset_pill_card`
      * Creates a multi page/tab UI with a singular `sidebar()` (which is
        shown on every page/tab).

    Parameters
    ----------
    *args
        Contents to the sidebar. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    width
        A valid CSS unit used for the width of the sidebar.
    position
        Where the sidebar should appear relative to the main content.
    open
        The initial state of the sidebar. It can be `"desktop"` (the sidebar starts open
        on desktop screen, closed on mobile), `"open"` or `True` (the sidebar starts
        open), `"closed"` or `False` (the sidebar starts closed), or `"always"` or
        `None` (the sidebar is always open and cannot be closed).

        In :func:`~shiny.experimental.ui.sidebar_toggle`, `open` indicates the desired
        state of the sidebar, where the default of `open = None` will cause the sidebar
        to be toggled open if closed or vice versa. Note that
        :func:`~shiny.experimental.ui.sidebar_toggle` can only open or close the
        sidebar, so it does not support the `"desktop"` and `"always"` options.
    id
        A character string. Required if wanting to re-actively read (or update) the
        `collapsible` state in a Shiny app.
    title
        A character title to be used as the sidebar title, which will be wrapped in a
        `<div>` element with class `sidebar-title`. You can also provide a custom
        :func:`~shiny.htmltools.tag` for the title element, in which case you'll
        likely want to give this element `class = "sidebar-title"`.
    bg,fg
        A background or foreground color.
    class_
        CSS classes for the sidebar container element, in addition to the fixed
        `.sidebar` class.
    max_height_mobile
        The maximum height of the horizontal sidebar when viewed on mobile devices.
        The default is `250px` unless the sidebar is included in a
        :func:`~shiny.experimental.ui.layout_sidebar` with a specified height, in
        which case the default is to take up no more than 50% of the layout container.

    Returns
    -------
    :
        A :class:`~shiny.experimental.ui.Sidebar` object.

    See Also
    --------
    * :func:`~shiny.experimental.ui.layout_sidebar`
    * :func:`~shiny.experimental.ui.navset_bar`
    * :func:`~shiny.experimental.ui.navset_tab_card`
    * :func:`~shiny.experimental.ui.navset_pill_card`
    """
    # TODO-future; validate `open`, bg, fg, class_, max_height_mobile

    if id is None and open != "always":
        # but always provide id when collapsible for accessibility reasons
        id = f"bslib-sidebar-{random.randint(1000, 10000)}"

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    if isinstance(title, (str, int, float)):
        title = div(str(title), class_="sidebar-title")

    collapse_tag = None
    if open != "always":
        collapse_tag = tags.button(
            _collapse_icon(),
            class_="collapse-toggle",
            type="button",
            title="Toggle sidebar",
            aria_expanded=trinary(open in ["open", "desktop"]),
            aria_controls=id,
        )

    tag = div(
        div(
            title,
            {"class": "sidebar-content"},
            *args,
        ),
        {"class": "bslib-sidebar-input"} if id is not None else None,
        {"class": "sidebar"},
        id=id,
        role="complementary",
        class_=class_,
    )

    return Sidebar(
        tag=tag,
        collapse_tag=collapse_tag,
        position=position,
        open=open,
        width=width,
        max_height_mobile=max_height_mobile,
        color_fg=fg,
        color_bg=bg,
    )


# TODO-maindocs; @add_example()
def layout_sidebar(
    sidebar: Sidebar | TagChild | TagAttrs,
    *args: TagChild | TagAttrs,
    fillable: bool = True,
    fill: bool = True,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    border: Optional[bool] = None,
    border_radius: Optional[bool] = None,
    border_color: Optional[str] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Sidebar layout

    Create a sidebar layout component which can be dropped inside any
    :func:`~shiny.ui.page` or :func:`~shiny.experimental.ui.card` context.

    Parameters
    ----------
    sidebar
        A :class:`~shiny.experimental.ui.Sidebar` object created by :func:`~shiny.experimental.ui.sidebar`.
    *args
        Contents to the main content area. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    fillable
        Whether or not the main content area should be wrapped in a fillable container.
        See :func:`~shiny.experimental.ui.as_fillable_container` for details.
    fill
        Whether or not the sidebar layout should be wrapped in a fillable container. See
        :func:`~shiny.experimental.ui.as_fill_item` for details.
    bg,fg
        A background or foreground color.
    border
        Whether or not to show a border around the sidebar layout.
    border_radius
        Whether or not to round the corners of the sidebar layout.
    border_color
        A border color.
    gap
        A CSS length unit defining the `gap` (i.e., spacing) between elements provided
        to `*args`. This argument is only applicable when `fillable = TRUE`.
    padding
        Padding to use for the body. This can be a numeric vector
        (which will be interpreted as pixels) or a character vector with valid CSS
        lengths. The length can be between one and four. If one, then that value
        will be used for all four sides. If two, then the first value will be used
        for the top and bottom, while the second value will be used for left and
        right. If three, then the first will be used for top, the second will be
        left and right, and the third will be bottom. If four, then the values will
        be interpreted as top, right, bottom, and left respectively.
    height
        Any valid CSS unit to use for the height.

    Returns
    -------
    :
        A :class:`~htmltools.Tag` object.

    See Also
    --------
    * :func:`~shiny.experimental.ui.sidebar`
    """
    updated_args = list(args)
    has_upgraded: bool = False
    # Use `args` here so `updated_args` can be safely altered in place
    for arg in args:
        if isinstance(arg, Sidebar):
            raise TypeError(
                "Please use the `sidebar=` argument to supply a `sidebar()`"
            )
        # TODO-future: >= 2023-11-01); Once `panel_sidebar()` is removed, we can remove this loop
        if isinstance(arg, DeprecatedPanelSidebar):
            if has_upgraded:
                raise TypeError(
                    "Multiple `panel_sidebar()` calls detected. Please use the `sidebar=` argument and supply a `sidebar()`"
                )
            if sidebar is not None:
                raise TypeError(
                    "A `panel_sidebar()` was supplied along with a `sidebar=` value. Please use only `sidebar=` to supply a `sidebar()`."
                )
            sidebar = arg.sidebar
            updated_args.remove(arg)
            has_upgraded = True

    if not isinstance(sidebar, Sidebar):
        sidebar = _sidebar_func(sidebar)

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    attrs, children = consolidate_attrs(*updated_args, **kwargs)
    # TODO-future: >= 2023-11-01); Once `panel_main()` is removed, we can remove this loop
    for child in children:
        if isinstance(child, DeprecatedPanelMain):
            attrs = consolidate_attrs(attrs, child.attrs)[0]
            # child.children will be handled when tagified

    main = div(
        {
            "role": "main",
            "class": f"main{' bslib-gap-spacing' if fillable else ''}",
            ""
            "style": css(
                background_color=bg,
                color=fg,
                gap=as_css_unit(gap),
                padding=as_css_padding(padding),
            ),
        },
        attrs,
        *children,
    )
    if fillable:
        main = as_fillable_container(main)

    max_height_mobile = sidebar.max_height_mobile or (
        "250px" if height is None else "50%"
    )

    res = div(
        {"class": "bslib-sidebar-layout"},
        {"class": "sidebar-right"} if sidebar.position == "right" else None,
        {"class": "sidebar-collapsed"} if sidebar.open == "closed" else None,
        main,
        sidebar.tag,
        sidebar.collapse_tag,
        sidebar_dependency(),
        _sidebar_init_js(),
        data_bslib_sidebar_init="true" if sidebar.open != "always" else None,
        data_bslib_sidebar_open=sidebar.open,
        data_bslib_sidebar_border=trinary(border),
        data_bslib_sidebar_border_radius=trinary(border_radius),
        style=css(
            __bslib_sidebar_width=as_css_unit(sidebar.width),
            __bslib_sidebar_bg=as_css_unit(sidebar.color_bg),
            __bslib_sidebar_fg=as_css_unit(sidebar.color_fg),
            __bs_card_border_color=border_color,
            height=as_css_unit(height),
            __bslib_sidebar_max_height_mobile=as_css_unit(max_height_mobile),
        ),
    )
    if fill:
        res = as_fill_item(res)

    return CardItem(res)


# TODO-maindocs; @add_example()
def sidebar_toggle(
    id: str,
    open: Literal["toggle", "open", "closed", "always"] | bool | None = None,
    session: Session | None = None,
) -> None:
    """
    Toggle a sidebar

    Toggle a :func:`~shiny.experimental.ui.sidebar` state during an active Shiny user session.

    Parameters
    ----------
    id
        The `id` of the :func:`~shiny.experimental.ui.sidebar` to toggle.
    open
        The desired state of the sidebar, choosing from the following options: `None`
        (toggle sidebar open and closed), `"open"` or `True` (open the sidebar),
        `"closed"` or `False` (close the sidebar). Note that `sidebar_toggle()` can only
        open or close the sidebar, so it does not support the `"desktop"` and `"always"`
    session
        A Shiny session object (the default should almost always be used).

    See Also
    --------
    * :func:`~shiny.experimental.ui.sidebar`
    * :func:`~shiny.experimental.ui.layout_sidebar`
    """
    session = require_active_session(session)

    method: Literal["toggle", "open", "close"]
    if open is None or open == "toggle":
        method = "toggle"
    elif open is True or open == "open":
        method = "open"
    elif open is False or open == "closed":
        method = "close"
    else:
        if open == "always" or open == "desktop":
            raise ValueError(
                f"`open = '{open}'` is not supported by `sidebar_toggle()`"
            )
        raise ValueError(
            "open must be NULL (or 'toggle'), TRUE (or 'open'), or FALSE (or 'closed')"
        )

    def callback() -> None:
        session.send_input_message(id, {"method": method})

    session.on_flush(callback, once=True)


_sidebar_func = sidebar


def _collapse_icon() -> Tag:
    return tags.svg(
        svgtags.path(
            fill_rule="evenodd",
            d="M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708z",
        ),
        xmlns="http://www.w3.org/2000/svg",
        viewBox="0 0 16 16",
        class_="bi bi-chevron-down collapse-icon",
        style="fill:currentColor;",
        aria_hidden="true",
        role="img",
    )


def _sidebar_init_js() -> Tag:
    # Note: if we want to avoid inline `<script>` tags in the future for
    # initialization code, we might be able to do so by turning the sidebar layout
    # container into a web component
    return tags.script(
        {"data-bslib-sidebar-init": True},
        "bslib.Sidebar.initCollapsibleAll()",
    )


########################################################


# TODO: use class to single out sidebar value (in layout_sidebar)
def panel_sidebar(
    *args: TagChild | TagAttrs,
    width: int = 4,
    **kwargs: TagAttrValue,
) -> DeprecatedPanelSidebar:
    """Deprecated. Please use :func:`~shiny.experimental.ui.sidebar` instead of
    `ui.panel_sidebar()`."""
    # TODO-future: >= 2023-11-01; Add deprecation message below
    # Plan of action:
    # * No deprecation messages today (2023-05-18), and existing code _just works_.
    # * Change all examples to use the new API.
    # * In, say, 6 months, start emitting messages for code that uses the old API.

    # warn_deprecated("Please use `sidebar()` instead of `panel_sidebar()`. `panel_sidebar()` will go away in a future version of Shiny.")
    return DeprecatedPanelSidebar(
        *args,
        width=width,
        **kwargs,
    )


class DeprecatedPanelSidebar:
    """
    [Deprecated] Sidebar panel

    Class returned from :func:`~shiny.experimental.ui.panel_sidebar`. Please do not
    use this class and instead supply your content to
    :func:`~shiny.experimental.ui.layout_sidebar` directly.

    Parameters
    ----------
    *args
        Contents to the sidebar. Or tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    width
        An integeger between 1 and 12, inclusive, that determines the width of the
        sidebar. The default is 4.
    **kwargs
        Tag attributes that are supplied to the resolved :class:`~htmltools.Tag` object.

    Attributes
    ----------
    sidebar
        A output from :func:`~shiny.experimental.ui.sidebar`.

    See Also
    --------
    * :func:`~shiny.experimental.ui.layout_sidebar`
    * :func:`~shiny.experimental.ui.sidebar`
    """

    # Store `attrs` for `layout_sidebar()` to retrieve
    sidebar: Sidebar

    def __init__(
        self, *args: TagChild | TagAttrs, width: int = 4, **kwargs: TagAttrValue
    ) -> None:
        self.sidebar = sidebar(
            *args,
            width=f"{int(width / 12 * 100)}%",
            **kwargs,
        )

    # Hopefully this is never used. But wanted to try to be safe
    def tagify(self) -> Tag:
        """
        Tagify the `self.sidebar.tag` and return the result in a TagList
        """
        return self.sidebar.tag.tagify()


def panel_main(
    *args: TagChild | TagAttrs,
    width: int = 8,
    **kwargs: TagAttrValue,
) -> TagList | DeprecatedPanelMain:
    """Deprecated. Please supply `panel_main(*args)` directly to `layout_sidebar()`."""
    # TODO-future: >= 2023-11-01; Add deprecation message below
    # warn_deprecated(
    #     "Please supply `panel_main(*args)` directly to `layout_sidebar()`."
    # )
    # warn if keys are being ignored
    attrs, children = consolidate_attrs(*args, **kwargs)
    if len(attrs) > 0:
        return DeprecatedPanelMain(attrs=attrs, children=children)
        warn_deprecated(
            "`*args: TagAttrs` or `**kwargs: TagAttrValue` values supplied to `panel_main()` are being ignored. Please supply them directly to `layout_sidebar()`."
        )

    return TagList(*children)


class DeprecatedPanelMain:
    """
    [Deprecated] Main panel

    Class returned from :func:`~shiny.experimental.ui.panel_main`. Please do not use
    this class and instead supply your content to
    :func:`~shiny.experimental.ui.layout_sidebar` directly.


    Parameters
    ----------
    attrs
        Attributes to apply to the parent tag of the children.
    children
        Children UI Elements to render inside the parent tag.

    Attributes
    ----------
    attrs
        Attributes to apply to the parent tag of the children.
    children
        Children UI Elements to render inside the parent tag.

    See Also
    --------
    * :func:`~shiny.experimental.ui.layout_sidebar`
    * :func:`~shiny.experimental.ui.sidebar`
    """

    # Store `attrs` for `layout_sidebar()` to retrieve
    attrs: TagAttrs
    # Return `children` in `layout_sidebar()` via `.tagify()` method
    children: list[TagChild]

    def __init__(self, *, attrs: TagAttrs, children: list[TagChild]) -> None:
        self.attrs = attrs
        self.children = children

    def tagify(self) -> TagList:
        """
        Tagify the `children` and return the result in a TagList
        """
        return TagList(self.children).tagify()
